/* 

    umesh2dgui, a GUI for umesh2d grid generator 

    Copyright (C) 2003 Gabriele Dini Ciacci
    Copyright (C) 2004 Gabriele Dini Ciacci
    Copyright (C) 2005 Gabriele Dini Ciacci
    Copyright (C) 2006 Gabriele Dini Ciacci
    Copyright (C) 2007 Gabriele Dini Ciacci
    Copyright (C) 2008 Gabriele Dini Ciacci

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "u2d_GLmesh.h"

#include <GL/gl.h>
#include <fltk/events.h>
#include <math.h>

#define BORDER 5
#define WHEEL_ZOOM_FACTOR -0.2
#define BUTTONPRESS_ZOOM_FACTOR 0.3

void u2d_GLmesh::draw() {
  if(!valid()) {
    /*projection
     *viewport
     *  w() and h() are the window dimenstions
     *  valid() is turned on after FLTK draw() returns
     */
    /*
    ortho();
    */
    set_default_colors();
    set_default_antialiasing();
    glViewport(BORDER, BORDER, this->w() - 2 * BORDER, this->h() - 2 * BORDER);
  }
  /* draw */
  MyInit();
  glClear(GL_COLOR_BUFFER_BIT);
  call_all_display_set();
  glFlush();
  //  swap_buffers();
}


int u2d_GLmesh::handle(int event) {
/* Don't call any OpenGL draw function from inside handle() */
  switch(event) {
    /* Those are events I want to to handle */
    case fltk::PUSH:
      if(fltk::event_button() == 1){ /* left button */
	last_push_x = fltk::event_x();
	last_push_y = fltk::event_y();
      }
      else if(fltk::event_button() == 2){ /* middle button */
	zoom(BUTTONPRESS_ZOOM_FACTOR, fltk::event_x(), fltk::event_y());
      }
      else if(fltk::event_button() == 3){ /* rigth button */
	zoom(-BUTTONPRESS_ZOOM_FACTOR, fltk::event_x(), fltk::event_y());
      }

      /* This say that we want to receive furter related events too*/
    return 1;
  case fltk::DRAG:
      if(fltk::event_button() == 1){
	pan((double)(last_push_x - fltk::event_x()), (double)(last_push_y - fltk::event_y()));
	last_push_x = fltk::event_x();
	last_push_y = fltk::event_y();
      }
    return 1;
  case fltk::RELEASE:
    return 1;
  case fltk::MOUSEWHEEL:
    centered_zoom(((double)fltk::event_dy() * WHEEL_ZOOM_FACTOR));
    return 1;
  default:
    /* The base class here handle the others */
    return fltk::GlWindow::handle(event);
  }
}

void u2d_GLmesh::pan(double disp_x, double disp_y){
  double new_minX;
  double new_maxX;
  double new_minY;
  double new_maxY;
  /* Calculate the new refence sizes as old +/- the displacement
   * converted from window pixel units to OpenGL normalized units.
   * NOTE on +/- :
   * the fact that x has + and y has - is due to different convetion
   * on the choice of the axis.
   * Event handler axis        |        OpenGL axis
   * (mouse coordinates)       |        (drawing coordinates)
   *                           |
   *   |                       |           ^
   * --+--------->             |           |
   *   |                       |           | 
   *   |                       |           |
   *   |                       |           |
   *   |                       |           |
   *   |                       |         --+---------->
   *   v                       |           |
   *
   * NOTE on normalization: 
   * We translate the reference sizes but normalize
   * on the real sizes, this is best behavor.
   */
  new_minX = ref_minX + disp_x * fabs(real_maxX - real_minX) / (this->w() - BORDER * 2);
  new_maxX = ref_maxX + disp_x * fabs(real_maxX - real_minX) / (this->w() - BORDER * 2);
  new_minY = ref_minY - disp_y * fabs(real_maxY - real_minY) / (this->h() - BORDER * 2);
  new_maxY = ref_maxY - disp_y * fabs(real_maxY - real_minY) / (this->h() - BORDER * 2);

#ifdef DEBUGGING
  printf("Displacement X: %f Y: %f\n", disp_x / (this->w() - BORDER * 2), \
	 disp_y / (this->h() - BORDER * 2));
  printf("real_size are: minX: %g, maxX: %g, minY: %g, MaxY: %g\n", real_minX, real_maxX, real_minY, real_maxY);
#endif
  set_ref_sizes(new_minX, new_maxX, new_minY, new_maxY);
  redraw();
}

void u2d_GLmesh::centered_zoom(double delta_factor){
  int click_X = (this->w() - BORDER * 2) / 2;
  int click_Y = (this->h() - BORDER * 2) / 2;
  zoom(delta_factor, click_X, click_Y);
}

void u2d_GLmesh::zoom(double delta_factor, int click_X, int click_Y){
  /* We get delta_factor cause it is of how much we want to zoom and not
   * the final factor we want to go, 0.1 is indeed 10% more not 1/10 of
   * the normal factor.
   */
  double new_minX;
  double new_maxX;
  double new_minY;
  double new_maxY;
  double length_X;
  double length_Y;
  double center_X;
  double center_Y;
  double visible_area;
  double zoomed_area;
  double aspect_ratio;
  /* We do an area zoom, not a length zoom */
  visible_area = fabs(real_maxX - real_minX) * fabs(real_maxY - real_minY);
  zoomed_area = visible_area * ( 1 - delta_factor );
#ifdef DEBUGGING
  printf("Delta factor: %f Visible area: %f, zoomed_area: %f\n",\
	 delta_factor, visible_area, zoomed_area);
#endif
  aspect_ratio =  fabs(real_maxX - real_minX) / fabs(real_maxY - real_minY);
  length_X = sqrt(zoomed_area * aspect_ratio);
  length_Y = length_X / aspect_ratio;
#ifdef DEBUGGING
  printf("length_X: %f length_Y: %f\n", length_X, length_Y);
#endif
  /* NOTE on +/-:
   * see pan() for explanation why we use +/-
   */
  center_X = real_minX + click_X * fabs(real_maxX - real_minX) / (this->w() - BORDER * 2);
  center_Y = real_maxY - click_Y * fabs(real_maxY - real_minY) / (this->h() - BORDER * 2);
#ifdef DEBUGGING
  printf("center_X: %f center_Y: %f\n", center_X, center_Y);
#endif
  new_minX = center_X - length_X / 2;
  new_maxX = center_X + length_X / 2;
  new_minY = center_Y - length_Y / 2;
  new_maxY = center_Y + length_Y / 2;
  set_ref_sizes(new_minX, new_maxX, new_minY, new_maxY);
  redraw();
}

/**
 * Obtain a valid display list id.
 * @return a valid display list id or 0 if cannot get a list
 * @author Gabriele Dini Ciacci <dark.schneider@iol.it>
 */
GLuint u2d_GLmesh::create_display_list(){
    GLuint display_list = glGenLists(1);
    if(display_list == 0) {
        fprintf(stderr, "Cannot get a dispaly list for OpenGL drawing\n");
        return 0;
    }
    if(glIsList(display_list) != GL_TRUE){
        printf("%i is NOT a valid list\n", display_list);
        return 0;
    }
#ifdef DEBUGGING
    printf("%i is a valid list\n", display_list);
#endif
    return display_list;
}

/**
 * Call glCallList on all display_set s.
 * If a given display_set has not a valid display list
 * it is just skipped.
 * @author Gabriele Dini Ciacci <dark.schneider@iol.it>
 */
void u2d_GLmesh::call_all_display_set(){
    vector<u2d_Display_Set>::iterator iter;
    for(iter = display_set.begin(); iter != display_set.end(); iter++) {
        if((*iter).display_list != 0) {
            glCallList((*iter).display_list);
        }
    }
}

/**
 * Call load_display_set on all display_set s.
 * If a given display_set is not loadable it just skips to the next.
 * @author Gabriele Dini Ciacci <dark.schneider@iol.it>
 */
void u2d_GLmesh::load_all_display_set(){
    vector<u2d_Display_Set>::iterator iter;
    for(iter = display_set.begin(); iter != display_set.end(); iter++) {
        load_display_set(&(*iter));
    }
}

/**
 * Load a display set. This means that the display list get compiled
 * based on the data that is in the set. If eveything goes fine the list
 * will be ready to be called with glCallList().
 * Check to use this in the future: EXT_compiled_vertex_array
 * @return 0 if the dispaly set wes loaded correctly.
 * @return -1 if there was a problem loading this display_set
 * @author Gabriele Dini Ciacci <dark.schneider@iol.it>
 */
int u2d_GLmesh::load_display_set(u2d_Display_Set *display_set) {
    if(display_set->display_list == 0) {
        display_set->display_list = create_display_list();
    }
    else {
#ifdef DEBUGGING
        printf("Deletting old list and recreating\n");
#endif 
        glDeleteLists(display_set->display_list, 1);
        display_set->display_list = create_display_list();
    }
    if(display_set->display_list == 0) {
        fprintf(stderr, "Cannot obtain a valid display list. Will not load this set\n");
        return -1;
    }
    /* Enable display lists */
    glEnableClientState(GL_VERTEX_ARRAY);
    /* Select the list we want to use and it's shape */
    glNewList(display_set->display_list, GL_COMPILE);
    if(display_set->Get_Vertex_Count() == 0) {
        fprintf(stderr, "No data to draw, vertex count is 0. Will not load this set\n");
        return -1;
    }
    glVertexPointer(display_set->Get_Coord_Per_Vertex(),\
                    GL_FLOAT, 0, (GLvoid *)display_set->Get_Vertex_Array());
    if(display_set->vertex_type == GL_LINES) {
        glDrawArrays(GL_LINES, 0, display_set->Get_Vertex_Count());
    }
    if(display_set->vertex_type ==  GL_TRIANGLES) {
        if(display_set->Get_Topology_Count() == 0) {
            fprintf(stderr, "No topology information for the given data. Will not load this set\n");
            return -1;
        }
        glDrawElements(GL_TRIANGLES, display_set->Get_Topology_Count(),\
                       GL_UNSIGNED_INT, display_set->Get_Topology_Array());
    }
    glEndList();
    glDisableClientState(GL_VERTEX_ARRAY);
}

void u2d_GLmesh::set_ref_sizes(double minX, double maxX, double minY, double maxY){
  /* This is a fix for:
   * When autoscaling the gl window, and only a one dimensional element,
   * like an horizontal or vertical line, is given the scale can't be
   * guessed correctly, that's because we do not have a reference
   * dimension for both X and Y so one of those will be 0 actually
   * preventing correct drawing.
   */

  if((minX != 0 || maxX != 0) && (minY == 0 && maxY == 0)){
#ifdef DEBUGGING
    printf("X size overriding Y\n");
#endif
    minY = minX;
    maxY = maxX;
  }
  else if((minY != 0 || maxY != 0) && (minX == 0 && maxX == 0)){
#ifdef DEBUGGING
    printf("Y size overriding X\n");
#endif
    minX = maxY;
    maxX = maxY;
  }
#ifdef DEBUGGING
  printf("ref_size sets: minX: %g, maxX: %g, minY: %g, MaxY: %g\n", minX, maxX, minY, maxY);
#endif
  ref_minX = minX;
  ref_maxX = maxX;
  ref_minY = minY;
  ref_maxY = maxY;
}

void u2d_GLmesh::get_ref_sizes(double *minX, double *maxX, double *minY, double *maxY){

  *minX = ref_minX;
  *maxX = ref_maxX;
  *minY = ref_minY;
  *maxY = ref_maxY;

}

void u2d_GLmesh::get_real_sizes(double *minX, double *maxX, double *minY, double *maxY){

  *minX = real_minX;
  *maxX = real_maxX;
  *minY = real_minY;
  *maxY = real_maxY;

}

void u2d_GLmesh::MyInit(){

  float factor;
  double view_minX;
  double view_maxX;
  double view_minY;
  double view_maxY;
  double center;
 
  /* This keep aspect ratio */

  double deltaY = ref_maxY - ref_minY;
  double deltaX = ref_maxX - ref_minX;
  if((deltaX / (double)this->w()) < (deltaY / (double)this->h())) {
    factor = (deltaY / (double)this->h()) / (deltaX / (double)this->w());
    center = (ref_maxX + ref_minX) / 2.0;
    view_minY = ref_minY;
    view_maxY = ref_maxY;
    view_minX = center - (center - ref_minX)*factor;
    view_maxX = center + (ref_maxX - center)*factor;
  }
  else {
    factor = (deltaX / (double)this->w()) / (deltaY / (double)this->h());
    center = (ref_maxY + ref_minY) / 2.0;
    view_minY = center - (center - ref_minY)*factor;
    view_maxY = center + (ref_maxY - center)*factor;
    view_minX = ref_minX;
    view_maxX = ref_maxX;
  }

  /* We take as reference the smaller area drawable */
  /* This algo does not keep aspect ratio */
  /*
  if(this->h() >= this->w()){
    factor = (double)this->h() / (double)this->w();
    center = (ref_maxY + ref_minY) / 2.0;
    view_minY = center - (center - ref_minY)*factor;
    view_maxY = center + (ref_maxY - center)*factor;
    view_minX = ref_minX;
    view_maxX = ref_maxX;
  }
  else{
    factor = (double)this->w() / (double)this->h();
    center = (ref_maxX + ref_minX) / 2.0;
    view_minY = ref_minY;
    view_maxY = ref_maxY;
    view_minX = center - (center - ref_minX)*factor;
    view_maxX = center + (ref_maxX - center)*factor;
  }
  */

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(view_minX * 1.00, view_maxX * 1.00, view_minY * 1.00, view_maxY * 1.00, -1, 1 );

  /* Set the real coordinates we are using */
  /* We have real sizes to imporve user experience while panning
   * and zomming, here below description of the bug fixed with real
   * sizes addition.
   * The fact is that due to one size beeing bigger than his actual
   * ref_size (to keep aspect ratio) mouse movement is not consistent
   * with actual pan, cause the figure "slide" under the mouse pointer
   * when panning along the biggest dimenstion.
   * Propsed fix:
   * Add a real_minX and so on, they get set in set_ref_sizes and use
   * them instead of ref_minX and so on.
   */
  real_minX = view_minX;
  real_maxX = view_maxX;
  real_minY = view_minY;
  real_maxY = view_maxY;
}

void u2d_GLmesh::set_default_colors(){
  /* Set up colors and line width */
  glClearColor(0.0, 0.0, 0.0, 0.0); //black
  glClear(GL_COLOR_BUFFER_BIT); // the background
  glColor3f(1.0, 1.0, 1.0);  //white
  glLineWidth((GLfloat)1); // 1 width for lines
}

void u2d_GLmesh::set_default_antialiasing(){
  /* The next 3 lines make opengl lines anti-aliased and good looking,
   * removing one will not do the trick anymore, look like they
   * work togheder
   */
  glEnable(GL_LINE_SMOOTH);
  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
}
