MODULE init_data
  
  USE grid_types  !  Grid types

  USE list        !  List operations

  USE curves      !  Splines Module
  
  USE delaunay    !  Delaunay basic module


IMPLICIT NONE

   !  ### 2D TOPOLOGICAL DEFINITIONS  ###

   TYPE vert
      TYPE (point), POINTER :: pnt
   END TYPE vert

   TYPE edge
      INTEGER :: index, ndata, begin_vert, end_vert, edge_kind, position
      INTEGER, DIMENSION(:)        , POINTER :: idata
      REAL (KIND=8), DIMENSION(:)  , POINTER :: sdata
      REAL (KIND=8), DIMENSION(:,:), POINTER :: hdata
      TYPE (point)                 , POINTER :: pnt_head, axp_head, add_head, &
                                             &  rem_head, gar_head      
      TYPE (simplex)               , POINTER :: spx_head, axs_head, int_head, &
                                             &  ext_head, stc_head
   END TYPE edge

   TYPE domain
      INTEGER :: number_of_verts, number_of_edges
      TYPE (vert), DIMENSION(:), POINTER :: verts
      TYPE (edge), DIMENSION(:), POINTER :: edges
      TYPE (point),   POINTER :: pnt_head, axp_head, add_head, rem_head, gar_head
      TYPE (simplex), POINTER :: spx_head, axs_head, int_head, ext_head, stc_head
   END TYPE domain

   INTEGER, PARAMETER, PUBLIC :: nd_b = 1, nd_d = 2

   TYPE (domain), PUBLIC :: grid, back
   PUBLIC  :: load_topology

CONTAINS

  SUBROUTINE init_boundary(idf1, idf2, idf3, name)

    !  routine to generate geometry.ANYNAME file starting from scratch
    !
    !  generates a global file containing geometric data, 
    !  as needed for build_grid

    !  idf1 : script file     logical unit  (edit.*)
    !  idf2 : output file     logical unit  (knots.*) 
    !  idf3 : data input file logical unit  (anyname.anyext)

    IMPLICIT NONE

    INTEGER, INTENT(IN) :: idf1, idf2, idf3
    CHARACTER (LEN=64)  :: name
    LOGICAL, PARAMETER  :: ascii = .TRUE.

    REAL (KIND=8), PARAMETER :: pi=3.14159265358979d0

    
    CHARACTER (LEN=20) :: command
    CHARACTER (LEN=60) :: infilename, outfilename, datafilename
    INTEGER ::iunit
    INTEGER :: n, m, i, j, k
    REAL (KIND=8) :: r1, r2, a1, a2, teta
    REAL (KIND=8), DIMENSION(2)   :: xb, xe, x0
    REAL (KIND=8), DIMENSION(2,2) :: rot_mat

    WRITE(*,*)
    WRITE(*,*)'Problem name ?'
    READ(*,*)name
    k = len_trim(name)

    iunit=idf1
    infilename = 'boundary.'//name(1:k)
    OPEN(idf1, file=infilename, status='old')
   
    WRITE(*,*)
    WRITE(*,*)'Loading boundary data:'
    READ (iunit,*)  ! #
    READ (iunit,*)  ! # GLOBAL DATA OF THE BOUNDARY GEOMETRY
    READ (iunit,*)  ! #
    READ (iunit,*)  ! NV, NE
    READ (iunit,*) grid % number_of_verts, grid % number_of_edges
    
    WRITE(*,*)' loading geometry...'
    READ (iunit,*)  ! #
    READ (iunit,*)  ! # GEOMETRY
    READ (iunit,*)  ! #
    
    m = grid % number_of_edges
    
    outfilename = 'knots.'//name(1:k)

    OPEN  (idf2, file = outfilename, form = 'formatted')

    WRITE (idf2,'(6x,a6)') 'N_CURV'
    WRITE (idf2,'(i12)') m

    DO i=1,m

       READ (iunit,*)  !BEGIN
       READ  (iunit,'(a20)') command  
       !by DD:to eliminate spaces before the real command
       k=0
       DO j=1,20        
        IF ( command(j:j).EQ.' ')THEN
          k=k+1
        ELSE
          EXIT
        ENDIF       
       ENDDO 
       DO j=1,20-k
          command(j:j)=command(j+k:j+k)
       ENDDO
       DO j=20-k,20
          command(j:j)=' '
       ENDDO
       !by DD:end
       
       SELECT CASE (command(1:3))

       CASE ('lin')

          READ  (iunit,*) n, xb, xe  !parameters
          CALL line   (idf2, n, xb, xe)

       CASE ('ell')

          READ  (iunit,*) n, xb, r1, r2, a1, a2
          a1 = pi*a1/180.d0
          a2 = pi*a2/180.d0

          CALL gen_ell(idf2, n, xb, r1, r2, a1, a2)

       CASE ('cir')

          READ  (iunit,*) n, xb, r1, a1, a2
          a1 = pi*a1/180.d0
          a2 = pi*a2/180.d0

          CALL gen_ell(idf2, n, xb, r1, r1, a1, a2)

       CASE ('dat')

          READ  (iunit,*) datafilename
          
          READ  (iunit,*)   !translation
          READ (iunit,*)x0
          
          READ  (iunit,*)  !rotation
          READ (iunit,*)teta
          teta = -teta*pi/180.d0

          OPEN  (idf3,file=datafilename,form='formatted', status = 'OLD')
          READ  (idf3,*) n  !number of datafile points
          CALL header (idf2, n, 0)
                     
          rot_mat(1,1) = COS(teta)
          rot_mat(2,1) = SIN(teta)
          rot_mat(1,2) = -rot_mat(2,1)
          rot_mat(2,2) = rot_mat(1,1)

          DO j=1,n
            
            READ  (idf3,*) xb  !point coordinates
            xe = MATMUL(rot_mat,(xb+x0))
            
            WRITE (idf2,'(2f12.5)') xe
          
          ENDDO


          CLOSE (idf3)

       CASE DEFAULT

          WRITE (*,*) 'command not known; no action taken'
          WRITE (*,*) 'the only command known are line, ellipse, circle, and data'
          STOP

       END SELECT
       READ(iunit,*)  ! END
       READ(iunit,*)  ! #
    ENDDO

    CLOSE (idf2)
    !WRITE(*,*) 
    !WRITE(*,*)'  created knots file: ', outfilename

    CALL make_geometry(idf2, name)
   
    CALL load_topology(idf1)
    
    CLOSE(idf1)

  CONTAINS
!--------------------------------------------------
    SUBROUTINE gen_ell(idf, n, c, r1, r2, beg, end)

      IMPLICIT NONE

      INTEGER, INTENT(IN) :: idf, n
      REAL (KIND=8), DIMENSION(2), INTENT(IN) :: c
      REAL (KIND=8), INTENT(IN) :: r1, r2, beg, end

      INTEGER :: i
      REAL (KIND=8) :: delta

      delta = (end-beg)/REAL(n-1,8)

      CALL header(idf, n, 2)

      WRITE(idf,'(2f20.10)') (c(1)+r1*cos(beg+delta*i),c(2)+   &
           r2*sin(beg+delta*i),i=0,n-1)
      
    END SUBROUTINE gen_ell
!-------------------------------------
    SUBROUTINE line (idf, n, beg, END)

      IMPLICIT NONE

      INTEGER, INTENT(IN) :: idf, n
      REAL (KIND=8), DIMENSION(2), INTENT(IN) :: beg, END

      INTEGER :: i
      REAL (KIND=8), DIMENSION(2) :: delta

      delta = (END-beg)/REAL(n-1,8)

      CALL header (idf, n, 0)

      WRITE (idf,'(2f20.10)') (beg+delta*i,i=0,n-1)

    END SUBROUTINE line
!--------------------------------
    SUBROUTINE header (idf, n, m)

      IMPLICIT NONE

      INTEGER, INTENT(IN) :: idf, n, m

      WRITE (idf,'(2a12)') '         DIM','      POINTS'
      WRITE (idf,'(2i12)') 2,n
      WRITE (idf,'( a12)') '         IDC'
      WRITE (idf,'( i12)') m
      WRITE (idf,'(2a12)') '           X','           Y'

    END SUBROUTINE header

  END SUBROUTINE init_boundary
!---------------------------------------
  SUBROUTINE make_geometry(idf, filename)

   !  Reads a geometrical data point file, computes
   !  the cubic spline passing through the nodes and
   !  save the results (hermite form).

   IMPLICIT NONE

    INTEGER           , INTENT(IN) :: idf
    CHARACTER (LEN=64), INTENT(IN) :: filename

    LOGICAL, PARAMETER :: ascii = .TRUE.

   INTEGER :: n

   n = len_trim(filename)

   OPEN  (idf,form='formatted',file='knots.'//filename(1:n))
   CALL new_curves  (idf, ascii)
   CLOSE (idf)

   OPEN  (idf,form='formatted',file='geometry.'//filename(1:n))
   CALL save_curves (idf, ascii)
   CLOSE (idf)

   !WRITE (*,*) '  created geometry file: geometry.',filename(1:n)
   !WRITE (*,*)
   RETURN

  END SUBROUTINE make_geometry

!

 SUBROUTINE load_topology(idf) 

 !  READ THE 2D TOPOLOGY AND INTIALIZE THE VARIOUS LISTS
      
 IMPLICIT NONE

 REAL(KIND=8) , PARAMETER     :: pi = 3.1415927464102
 INTEGER      , INTENT(IN)    :: idf

 INTEGER :: i, j, n

 IF ( .NOT. ASSOCIATED ( grid % verts ) ) THEN
	 ALLOCATE ( grid % verts(grid % number_of_verts) )
 ENDIF

 IF ( .NOT. ASSOCIATED ( grid % edges ) ) THEN
	 ALLOCATE ( grid % edges(grid % number_of_edges) )
 ENDIF

 WRITE(*,*)' loading topology...'
 READ(idf,*) ! TOPOLOGY
 
 DO i=1,grid % number_of_edges

   READ (idf,*) ! #
   READ (idf,*) ! BEGIN
   READ (idf,*) ! INDEX, NDATA, BEGIN_VERT, END_VERT
   READ (idf,*) grid % edges(i) % index,         &
	            grid % edges(i) % ndata,         &
	            grid % edges(i) % begin_vert,    &
	            grid % edges(i) % end_vert
               
   n = grid % edges(i) % ndata
  
  !by DD begin   
   grid % edges(i) % edge_kind = 0
   grid % edges(i) % position = 0
  !by DD end	 
   IF ( n < 2 ) THEN
	   WRITE (*,*) 'load_topology error, valid ndata >= 2; I have ',n
	   WRITE (*,*) '(stop)'
	   STOP
   ENDIF

   ALLOCATE ( grid % edges(i) % idata(n) )
	 ALLOCATE ( grid % edges(i) % sdata(n) )
	 ALLOCATE ( grid % edges(i) % hdata(2*nd_d-1,n) )

   READ (idf,*) ! IDATA, SDATA, HDATA
	 DO j=1,n
	    READ (idf,*) grid % edges(i) % idata(j),   &
			 grid % edges(i) % sdata(j),   &
			 !by DD begin
       grid % edges(i) % hdata(1,j)
       grid%edges(i)%hdata(2,j) = grid%edges(i)%hdata(1,j)
       grid%edges(i)%hdata(3,j) = zero
       !by DD end
	 ENDDO
   READ (idf,*) ! END

   grid % edges(i) % pnt_head  =>  newhead_point ()
   grid % edges(i) % axp_head  =>  newhead_point ()
   grid % edges(i) % add_head  =>  newhead_point ()
   grid % edges(i) % rem_head  =>  newhead_point ()
   grid % edges(i) % gar_head  =>  newhead_point ()

   grid % edges(i) % spx_head  =>  newhead_simplex ()
   grid % edges(i) % axs_head  =>  newhead_simplex ()
   grid % edges(i) % int_head  =>  newhead_simplex ()
   grid % edges(i) % ext_head  =>  newhead_simplex ()
   grid % edges(i) % stc_head  =>  newhead_simplex ()

 ENDDO

 grid % pnt_head  =>  newhead_point ()
 grid % axp_head  =>  newhead_point ()
 grid % add_head  =>  newhead_point ()
 grid % rem_head  =>  newhead_point ()
 grid % gar_head  =>  newhead_point ()

 grid % spx_head  =>  newhead_simplex ()
 grid % axs_head  =>  newhead_simplex ()
 grid % int_head  =>  newhead_simplex ()
 grid % ext_head  =>  newhead_simplex ()
 grid % stc_head  =>  newhead_simplex ()


END SUBROUTINE load_topology


END module init_data
