/*   umesh2dgui, a GUI for umesh2d grid generator 

    Copyright (C) 2003 Gabriele Dini Ciacci
    Copyright (C) 2004 Gabriele Dini Ciacci
    Copyright (C) 2005 Gabriele Dini Ciacci
    Copyright (C) 2006 Gabriele Dini Ciacci
    Copyright (C) 2007 Gabriele Dini Ciacci
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
/* included for realpath() */
#include <limits.h>
/* for getcwd() */
#include <unistd.h>
#include <errno.h>
/* for stat() */
#include <sys/types.h>
#include <sys/stat.h>


#include <fltk/xpmImage.h>
#include <fltk/Browser.h>
#include <fltk/MultiBrowser.h>
#include <fltk/run.h>
#include <fltk/Button.h>
#include <fltk/Window.h>
#include <fltk/Group.h>
#include <fltk/CheckButton.h>
#include <fltk/Output.h>
#include <fltk/Color.h>
#include <fltk/filename.h>
#include <fltk/RadioButton.h>
#include <fltk/RadioItem.h>
#include <fltk/TextDisplay.h>

#include "umesh2dgui.h"
#include "u2d_core.h"
#include "u2d_GL.h"
#include "BDFile.h"
#include "ascii_color.h"
#include "u2d_GLmesh.h"
#include "u2d_load_plotmtv.h"

#include "images/open_icon.xpm"
#include "images/save_7.xpm"
#include "images/cancel_15.xpm"
#include "images/create_12.xpm"


#define COLUMN1 50
#define COLUMN2 190
#define COLUMN3 310
#define COLUMN4 450

#define BROWSER_W 640
#define BROWSER_H 240
#define GLVIEW_W 202
#define GLVIEW_H 202
#define TITLEBAR_OFFSET 35
#define MANAGER_BORDER 5

#define MSGBUF_LENGTH 1024

/* Fixed for now, to move in configure */
#define USE_PNG 1

#ifdef USE_PNG 
#include <fltk/SharedImage.h>
#define LOAD_ICON_IMAGE(img) fltk::pngImage::get(img)
#else
/* Remember to include here the xpm image */
#define LOAD_ICON_IMAGE(img) fltk::xpmImage(img)
#endif

/* GLOBALS */
u2d_GL *u2d_gl;
u2d_GLmesh *u2d_glmesh;
BDFile *bdfile;
u2d_Core *u2d_core;
/* the tabs */
fltk::Group *boundarygroup;
fltk::Group *domaingroup;
fltk::Group *creationgroup;
fltk::Group *outputgroup;
fltk::Group *meshgroup;
/* useful groups */
fltk::ScrollGroup *datagroup;
fltk::ScrollGroup *steinergroup;
fltk::ScrollGroup *backgridgroup;
fltk::MultiBrowser *prim_browser;
/* the three windows */
fltk::Window *window;
fltk::Window *opengl_window;
fltk::Window *prim_browser_window;
/* the correspoding useful group subwindows */
fltk::Group *boundary_subwindow;
fltk::Group *creation_subwindow;
fltk::Group *domain_subwindow;
/* the message bar */
fltk::Output *msgbar;
/* the text area for shell output */
fltk::TextDisplay *shell_output;
/* the primitive menubar items */
fltk::Item *lineitem;
fltk::Item *circleitem;
fltk::Item *ellipseitem;
fltk::Item *dataitem;
fltk::Item *rectangleitem;
/* a terrible list of indexes for primitive values*/
int inputindex[10];
int filenameinput_i; //trailing _i is for index
/* integer that keep track of scroll position */
int ydataposition = 0;
int ysteinerposition = 0;
int ybackgridposition = 0;
/* the file format requested, to be moved to class when redesigning */
int fileformat = FORMAT_UMESH2D;
/* global icons */
fltk::xpmImage *cancel_icon;
/* a buffer for the message bar */
char msgbuf[MSGBUF_LENGTH];

int main(int argc, char ** argv){

  /* PRINT SOME ABOUT INFO TO STDOUT */
  fprintf(stdout, "%s%s%s by %sGabriele Dini Ciacci%s - version %s%s%s\n", \
  ASC_RED, PACKAGE, ASC_RESET, ASC_GREEN, ASC_RESET, ASC_BLUE, VERSION, ASC_RESET);

  /* PARSE THE ARGUMENTS WITH FLTK INTERNAL FUNCTION */
  fltk::args(argc, argv);

  /* CLASS DECLARATION */
  bdfile = new BDFile;
  u2d_core = new u2d_Core;
  window = new fltk::Window(640, 534, "umesh2dgui");
  opengl_window = new fltk::Window(GLVIEW_W, GLVIEW_H, "OpenGL preview");
  prim_browser_window = new fltk::Window(BROWSER_W, BROWSER_H, "Primitive browser");

  /* ICON IMAGE DECLATATION */
  fltk::xpmImage *open_icon;
  open_icon = new fltk::xpmImage(open_icon_xpm);
  fltk::xpmImage *save_icon;
  save_icon = new fltk::xpmImage(save_7_xpm);
  fltk::xpmImage *create_icon;
  create_icon = new fltk::xpmImage(create_12_xpm);
  /* cancel_icon is declared as a global */
  cancel_icon = new fltk::xpmImage(cancel_15_xpm);

  /* WINDOW SETUP */
  window->callback((fltk::Callback *)quititem_cb);
  window->begin();
  fltk::MenuBar *menubar = new fltk::MenuBar(0, 0, 640, 30, 0);
  fltk::TabGroup *tabbar = new fltk::TabGroup(10, 30, 620, 480, 0);
  msgbar = new fltk::Output(0, 510, 640, 24);
  //  msgbar->deactivate();
  msgbar->color(fltk::GRAY75);
  msgout("Wellcome to the creation wizard! Select a primitive to start");
  window->end();

  /* OPENGL WINDOW SETUP*/
  opengl_window->begin();
  u2d_gl = new u2d_GL::u2d_GL(1, 1, GLVIEW_W - 2, GLVIEW_H - 2, 0);
  opengl_window->end();
  opengl_window->resizable(u2d_gl);

  /* PRIMITIVE BROWSER WINDOW SETUP */
  const char *prim_browser_labels[] = {
    "Id",
    "Primitive Name",
    "Obj",
    "Type",
    "x1",
    "y1",
    "x2",
    "y2",
    "radius",
    "S arc",
    "E arc",
    0
  };
  int prim_browser_widths[] = {
    20,
    -1,
    20,
    50,
    70,
    70,
    70,
    70,
    70,
    50,
    50,
    0
  };
  prim_browser_window->begin();
  prim_browser = new fltk::MultiBrowser(0, 0, prim_browser_window->w(), \
					prim_browser_window->h() - 45, \
					"Primitive List");
  prim_browser->column_widths(prim_browser_widths);
  prim_browser->column_labels(prim_browser_labels);

  fltk::Button *browser_del_prim_button = new fltk::Button(10, prim_browser_window->h() - 35, 70, 25, "Delete");
  browser_del_prim_button->callback((fltk::Callback *)browser_del_prim_cb);
  /* Name button */
  /*
  fltk::Button *prim_list_name;
  prim_list_name = new fltk::Button(1, 1, 200, 20, "Primitive Name");
  prim_list_name->align(fltk::ALIGN_LEFT | fltk::ALIGN_INSIDE);
  */
  /* Type button */
  /*
  fltk::Button *prim_list_type;
  prim_list_type = new fltk::Button(200, 1, 50, 20, "Type");
  prim_list_type->align(fltk::ALIGN_LEFT | fltk::ALIGN_INSIDE);
  */
  /* x1 button */
  /*
  fltk::Button *prim_list_x1;
  prim_list_x1 = new fltk::Button(250, 1, 50, 20, "x1 coord");
  prim_list_x1->align(fltk::ALIGN_LEFT | fltk::ALIGN_INSIDE);
  */
  /* y1 button */
  /*
  fltk::Button *prim_list_y1;
  prim_list_y1 = new fltk::Button(300, 1, 50, 20, "y1 coord");
  prim_list_y1->align(fltk::ALIGN_LEFT | fltk::ALIGN_INSIDE);
  */
  /* x2 button */
  /*
  fltk::Button *prim_list_x2;
  prim_list_x2 = new fltk::Button(350, 1, 50, 20, "x2 coord");
  prim_list_x2->align(fltk::ALIGN_LEFT | fltk::ALIGN_INSIDE);
  */
  /* y2 button */
  /*
  fltk::Button *prim_list_y2;
  prim_list_y2 = new fltk::Button(400, 1, 50, 20, "y2 coord");
  prim_list_y2->align(fltk::ALIGN_LEFT | fltk::ALIGN_INSIDE);
  */

  prim_browser_window->end();

  /* TAB BAR SETUP */
  //  tabbar->box(fltk::UP_BOX);
  //  tabbar->box(fltk::BORDER_BOX);
  tabbar->box(fltk::THIN_UP_BOX);
  tabbar->begin(); // 620 450
  boundarygroup = new fltk::Group(0, 30, 620, 440, "Boundary");
  boundarygroup->box(fltk::ENGRAVED_BOX);
  domaingroup =  new fltk::Group(0, 30, 620, 440, "Domain");
  domaingroup->box(fltk::EMBOSSED_BOX);
  creationgroup = new fltk::Group(0, 30, 620, 440, "Creation");
  creationgroup->box(fltk::EMBOSSED_BOX);
  outputgroup = new fltk::Group(0, 30, 620, 440, "Output");
  outputgroup->box(fltk::EMBOSSED_BOX);
  meshgroup = new fltk::Group(0, 30, 620, 440, "Mesh");
  meshgroup->box(fltk::EMBOSSED_BOX);
  tabbar->end();

  /* TITLE BOX SETUP */ /* removed now */
  /*  fltk::Widget *box = new fltk::Widget(10, 30, 620, 40, "umesh2dgui");
   *  box->box(fltk::UP_BOX);
   *  box->labelfont(fltk::HELVETICA_BOLD_ITALIC);
   *  box->labelcolor(fltk::GREEN);
   *  box->labelsize(40);
   *  box->labeltype(fltk::SHADOW_LABEL);
   */

  /* BOUNDARY TAB */
  boundarygroup->begin();

  /* next (and prev) button */
  fltk::Button *boundary_nextbutton = new fltk::Button(320, 400, 100, 25, "Next  ");
  boundary_nextbutton->callback((fltk::Callback *)boundary_nextbutton_cb);
  boundary_nextbutton->image(LOAD_ICON_IMAGE("images/next.png"));
  boundary_nextbutton->align(fltk::ALIGN_RIGHT | fltk::ALIGN_INSIDE);
  boundary_nextbutton->tooltip("Proceed to domain definition");



  /* the subwindow */
  //  boundary_subwindow = new fltk::Group(10, 50, 600, 340, "Insert primitive data here:");
  boundary_subwindow = new fltk::Group(10, 50, 600, 340, "");
  boundary_subwindow->box(fltk::SHADOW_BOX);



  /* PRIMITIVE MENU SETUP */
  /*  this is inside boundary tab */

  /* Choice */
  /*
   *  fltk::Choice *primitivemenu = new fltk::Choice(210, 10, 100, 20,\
   *    "Choose type of primitive to add:");
   *  primitivemenu->add("&clear", 0, (fltk::Callback *)clear_cb, 0, 0);
   *  primitivemenu->add("&line", 0, (fltk::Callback *)line_cb, 0, 0);
   *  primitivemenu->add("&circle", 0, (fltk::Callback *)circle_cb, 0, 0);
   *  primitivemenu->add("&ellipse", 0, (fltk::Callback *)ellipse_cb, 0, 0);
   *  primitivemenu->add("&data", 0, (fltk::Callback *)clear_cb, 0, 0);
   */

  fltk::MenuBar *primitivemenu = new fltk::MenuBar(5, 2, 600, 30); 

  primitivemenu->begin();
  lineitem = new fltk::Item("Line");
  lineitem->callback((fltk::Callback*)line_cb);
  lineitem->image(LOAD_ICON_IMAGE("images/line_1.png"));
  circleitem = new fltk::Item("Circle");
  circleitem->callback((fltk::Callback*)circle_cb);
  circleitem->image(LOAD_ICON_IMAGE("images/circle_1.png"));
  ellipseitem = new fltk::Item("Ellipse");
  ellipseitem->callback((fltk::Callback*)ellipse_cb);
  ellipseitem->image(LOAD_ICON_IMAGE("images/ellipse_1.png"));
  dataitem = new fltk::Item("Data");
  dataitem->callback((fltk::Callback*)data_cb);
  dataitem->image(LOAD_ICON_IMAGE("images/data_1.png"));
  rectangleitem = new fltk::Item("Rectangle");
  rectangleitem->callback((fltk::Callback*)rectangle_cb);
  rectangleitem->image(LOAD_ICON_IMAGE("images/rectangle_2.png"));
  primitivemenu->end();

  boundarygroup->end();

  /* DOMAIN TAB */
  domaingroup->begin();

  /* next and prev buttons */
  fltk::Button *domain_nextbutton = new fltk::Button(320, 400, 100, 25, "Next  ");
  domain_nextbutton->callback((fltk::Callback *)domain_nextbutton_cb);
  domain_nextbutton->image(LOAD_ICON_IMAGE("images/next.png"));
  domain_nextbutton->align(fltk::ALIGN_RIGHT | fltk::ALIGN_INSIDE);
  domain_nextbutton->tooltip("Proceed to file output");

  fltk::Button *domain_prevbutton = new fltk::Button(200, 400, 100, 25, "  Prev");
  domain_prevbutton->callback((fltk::Callback *)domain_prevbutton_cb);
  domain_prevbutton->image(LOAD_ICON_IMAGE("images/prev.png"));
  domain_prevbutton->align(fltk::ALIGN_LEFT | fltk::ALIGN_INSIDE);
  domain_prevbutton->tooltip("Back to boundary definition");

  domain_subwindow = new fltk::Group(10, 50, 600, 340, "");
  domain_subwindow->box(fltk::SHADOW_BOX);

  /* the subwindow */
  domain_subwindow->begin();

  /* backgrid bordered box*/
  fltk::Group *backgridcontour  = new fltk::Group(10, 10, 580, 145, "");
  backgridcontour->box(fltk::BORDER_FRAME);
  backgridcontour->align(fltk::ALIGN_INSIDE_LEFT);

  backgridcontour->begin();

  fltk::Button *new_backgrid = new fltk::Button(450, 115, 120, 20, "Add backgrid point");
  new_backgrid->callback((fltk::Callback *)new_backgrid_cb);

  backgridgroup = new fltk::ScrollGroup(10, 10, 560, 100, "");
  backgridgroup->scrollbar.linesize(20);

  backgridcontour->end();

  /* steiner bordered box*/
  fltk::Group *steinercontour  = new fltk::Group(10, 170, 580, 145, "");
  steinercontour->box(fltk::BORDER_FRAME);
  steinercontour->align(fltk::ALIGN_INSIDE_LEFT);

  steinercontour->begin();

  fltk::Button *new_steiner = new fltk::Button(450, 115, 120, 20, "Add steiner point");
  new_steiner->callback((fltk::Callback *)new_steiner_cb);

  steinergroup = new fltk::ScrollGroup(10, 10, 560, 100, "");
  steinergroup->scrollbar.linesize(20);

  steinercontour->end();

  domain_subwindow->end();

  domaingroup->end();


  /* CREATION TAB */
  creationgroup->begin();

  /* next and prev buttons */
  fltk::Button *creation_prevbutton = new fltk::Button(200, 400, 100, 25, "  Prev");
  creation_prevbutton->callback((fltk::Callback *)creation_prevbutton_cb);
  creation_prevbutton->image(LOAD_ICON_IMAGE("images/prev.png"));
  creation_prevbutton->align(fltk::ALIGN_LEFT | fltk::ALIGN_INSIDE);
  creation_prevbutton->tooltip("Back to domain definition");

  fltk::Button *creation_nextbutton = new fltk::Button(320, 400, 100, 25, "Finish  ");
  creation_nextbutton->callback((fltk::Callback *)creation_nextbutton_cb);
  creation_nextbutton->image((fltk::Image *)create_icon);
  creation_nextbutton->align(fltk::ALIGN_RIGHT | fltk::ALIGN_INSIDE);
  creation_nextbutton->tooltip("Output file and generate mesh");

  creation_subwindow = new fltk::Group(10, 50, 600, 340, "");
  creation_subwindow->box(fltk::SHADOW_BOX);

  creation_subwindow->begin();

  /* output format bordered box */
  fltk::Group *outcontour  = new fltk::Group(COLUMN1, 31, 240, 67, "Select output\n     format");
  outcontour->box(fltk::BORDER_FRAME);
  outcontour->align(fltk::ALIGN_INSIDE_LEFT);

  outcontour->begin();
  fltk::RadioButton *umesh2d_format = new fltk::RadioButton(130, 3, 98, 20, "umesh2d");
  umesh2d_format->callback((fltk::Callback *)umesh2d_format_cb);
  umesh2d_format->value(true);
  fltk::RadioButton *umesh2d_fe_format = new fltk::RadioButton(130, 21, 98, 20, "umesh2d FE");
  umesh2d_fe_format->callback((fltk::Callback *)umesh2d_fe_format_cb);
  fltk::RadioButton *umesh2d_fv_format = new fltk::RadioButton(130, 39, 98, 20, "umesh2d FV");
  umesh2d_fv_format->callback((fltk::Callback *)umesh2d_fv_format_cb);
  outcontour->end();


  /* Filename */
  fltk::Input *filenameinput = new fltk::Input(110, 300, 250, 25, "Project name :");
  //  filenameinput->box(fltk::SHADOW_BOX);
  filenameinput_i = creation_subwindow->find(filenameinput);

  /* Filename browser button */
  fltk::Button *filenamebrowser = new fltk::Button(362, 300, 25, 25, "");
  filenamebrowser->callback((fltk::Callback *)openfilebrowser);
  filenamebrowser->image((fltk::Image *)open_icon);
  filenamebrowser->tooltip("Browse");
  //  filenamebrowser->buttonbox(fltk::SHADOW_BOX);

  creation_subwindow->end();

  creationgroup->end();

  /* OUTPUT TAB */
  outputgroup->begin();
  shell_output = new fltk::TextDisplay(1, 1, 620, 440);
  outputgroup->end();

  /* MESH TAB */
  meshgroup->begin();

  /* opengl mesh window setup */
  u2d_glmesh = new u2d_GLmesh::u2d_GLmesh(1, 1, 620, 440, 0);
  meshgroup->end();

  /* MENUBAR SETUP */
  menubar->begin();
  /* file menu */
  fltk::ItemGroup *filemenu = new fltk::ItemGroup("&File");
  filemenu->begin();
  fltk::Item *openitem = new fltk::Item("&Open");
  openitem->callback((fltk::Callback *)openitem_cb);
  fltk::Item *saveitem = new fltk::Item("&Save");
  saveitem->callback((fltk::Callback *)saveitem_cb);
  saveitem->image((fltk::Image *)save_icon);
  //  cancel->tooltip("Cancel");
  fltk::Item *saveasitem = new fltk::Item("&Save As...");
  saveasitem->callback((fltk::Callback *)saveasitem_cb);
  fltk::Item *quititem = new fltk::Item("&Quit");
  quititem->callback((fltk::Callback *)quititem_cb);
  filemenu->end();
  /* view menu */
  fltk::ItemGroup *view = new fltk::ItemGroup("&View");
  view->begin();
  fltk::Item *glview = new fltk::Item("&Toggle openGL preview");
  glview->callback((fltk::Callback *)glview_cb);
  glview->type(fltk::Button::TOGGLE);
  fltk::Item *browserview = new fltk::Item("&Toggle Primirive browser");
  browserview->callback((fltk::Callback *)browserview_cb);
  browserview->type(fltk::Button::TOGGLE);
  view->end();
  /* tools menu */
  fltk::ItemGroup *toolsmenu = new fltk::ItemGroup("&Tools");
  toolsmenu->begin();
  fltk::Item *main_coreitem = new fltk::Item("&Evaluate closed objects");
  main_coreitem->callback((fltk::Callback *)main_core_cb);
  toolsmenu->end();
  /* help menu */
  fltk::ItemGroup *help = new fltk::ItemGroup("&Help");
  help->begin();
  fltk::Item *about = new fltk::Item("&About umesh2dgui");
  about->callback((fltk::Callback *)about_cb);

#ifdef DEBUGGING  
  fltk::Item *debugitem = new fltk::Item("&Debug callback");
  debugitem->callback((fltk::Callback *)debug_cb);
#endif
  
  help->end();
  menubar->end();

  /* set focus on first next button */
  //  fltk::focus(boundary_nextbutton);

  /* RESIZABLE SETUP */
  prim_browser_window->resizable(prim_browser_window);


  /* LOOP*/
  window->show(argc, argv);
  boundary_subwindow->show();
  line_cb();

  opengl_window->show();

  prim_browser_window->show();

  int val;

  val =  fltk::wait();

  /* Now that we got from system the window position we
   * reposition out windows in a nice way */
  prim_browser_window->position(window->x(), window->y() + window->h() + TITLEBAR_OFFSET);
  //  prim_browser_window->position(window->x(), window->y() - BROWSER_H - TITLEBAR_OFFSET);
  opengl_window->position(window->x() + window->w() + MANAGER_BORDER, window->y());

  val = fltk::run();

  /* clear the childs */
  window->clear();
  opengl_window->clear();
  prim_browser_window->clear();

  /* deinstantiate the parents */
  delete window;
  delete opengl_window;
  delete prim_browser_window;
  /* delete the icons */
  delete open_icon;
  delete save_icon;
  delete create_icon;
  delete cancel_icon;
  /* delete our classes */
  delete bdfile;
  delete u2d_core;

  return val;

}

void openfilebrowser() {
  ((fltk::Input *)creation_subwindow->child(filenameinput_i))->value(fltk::file_chooser("Select a file", NULL, NULL, true)); 
  /* try to call file browser several time. Value() wants a good static buffer or what?
   * LOOK LIKE WORKING... but i'm really not sure if will under extreme condition
   */
}

void creation_nextbutton_cb() {
  int returnvalue;
  if(strlen(((fltk::Input *)creation_subwindow->child(filenameinput_i))->value()) < 1){
    printf("Filename is empty. Opening file browser\n");
    openfilebrowser();
  }
  if(strlen(((fltk::Input *)creation_subwindow->child(filenameinput_i))->value()) < 1){
    printf("Filename is empty. User aborted\n");
    return;
  }

  switch(fileformat) {
  case FORMAT_UMESH2D :
    bdfile->set_output_format(FORMAT_UMESH2D);
    break;
  case FORMAT_UMESH2D_FE :
    bdfile->set_output_format(FORMAT_UMESH2D_FE);
    break;
  case FORMAT_UMESH2D_FV :
    bdfile->set_output_format(FORMAT_UMESH2D_FV);
    break;
  default : {
    fprintf(stderr, "File format not supported\n");
    msgout("File format not supported");
    break;
  }
  }

  /* Copy steiner and backgrid info to u2d_core*/
  u2d_core->allocate_steiner(steinergroup->children() / 3);
#ifdef DEBUGGING
  printf("Steiners in widgets: %i\n", steinergroup->children() / 3);
#endif
  for(int i = 0; i < steinergroup->children(); i += 3){
    u2d_core->set_steiner(i / 3, ((fltk::ValueInput *)steinergroup->child(i))->value(),
			  ((fltk::ValueInput *)steinergroup->child(i + 1))->value());
  }

  u2d_core->allocate_backgrid(backgridgroup->children() / 4);
#ifdef DEBUGGING
  printf("Backgrid in widgets: %i\n", backgridgroup->children() / 4);
#endif
  for(int i = 0; i < backgridgroup->children(); i += 4){
    u2d_core->set_backgrid(i / 4, ((fltk::ValueInput *)backgridgroup->child(i))->value(),
			   ((fltk::ValueInput *)backgridgroup->child(i + 1))->value(),
			   ((fltk::ValueInput *)backgridgroup->child(i + 2))->value());
  }

  /* Set filenames in bdfile */
  bdfile->set_project_name(((fltk::Input *)creation_subwindow->child(filenameinput_i))->value());

  /* Calculate closed objects */
  fprintf(stdout, "Calculating closed objects and vertex enumeration...\n");
  u2d_core->find_objects();
  fprintf(stdout, "Done!\n");

  /* Initializate the BDFile class with necessary
   * parameters */
  bdfile->select_prim_list(u2d_core->get_prim_list_start(), u2d_core->get_prim_list_end(), \
			   u2d_core->get_prim_number(), u2d_core->get_vertex_n());
  bdfile->set_steiners(u2d_core->steiners, u2d_core->steiner_n);
  bdfile->set_backgrids(u2d_core->backgrids, u2d_core->backgrid_n);


  /* Call the output boundary file function of the
   * selected class for output */
  returnvalue = bdfile->outputbfile();
  if(returnvalue < 0) {
    fprintf(stderr, "createfile_cb: Error creating the boundary file\n");
    if(returnvalue == -1) fltk::alert("Error creating boundary file: Invalid filename\n");
    else if(returnvalue == -2) fltk::alert("Error creating boundary file: File opening was not possible\n");
    else if(returnvalue == -3) fltk::alert("Notting to output or list initialization error\n");
    else fltk::alert("Error creating boundary file: Unkown reason\n");
    return;
  }
  returnvalue = bdfile->outputdfile();
  if(returnvalue < 0) {
    fprintf(stderr, "createfile_cb: Error creating the domain file\n");
    if(returnvalue == -1) fltk::alert("Error creating domain file: Invalid filename\n");
    else if(returnvalue == -2) fltk::alert("Error creating domain file: File opening was not possible\n");
    else if(returnvalue == -3) fltk::alert("Notting to output or list initialization error\n");
    else fltk::alert("Error creating domain file: Unkown reason\n");
    return;
  }
  fprintf(stdout, "File creation sucessfull\n");
  msgout("File creation sucessfull");

  /* Show the output tab */
  creationgroup->hide();
  outputgroup->show();

  /* Run the mesher */
  execute_mesher();
  draw_mesh();

  /* Show the mesh tab */
  outputgroup->hide();
  meshgroup->show();
}



void line_cb(){
  /* POPULATE THE BOUNDARY_SUBWINDOW */
  clear_cb();
  lineitem->box(fltk::DOWN_BOX);
  boundary_subwindow->begin();

  /* spline*/
  fltk::ValueInput *splinepoints = new fltk::ValueInput(COLUMN2, 10, 100, 20, "N. points for spline aprox.: ");
  splinepoints->range(0, 10000);
  splinepoints->linesize(1.0);
  splinepoints->step(1);
  splinepoints->set_value(100);
  //  splinepoints->pagesize(10.0);
  inputindex[0] = boundary_subwindow->find(splinepoints);
  //  printf("%i", inputindex[0]);

  /* name */
  fltk::Input *primname = new fltk::Input(COLUMN4, 10, 100, 20, "Name (Optional): ");
  inputindex[1] = boundary_subwindow->find(primname);

  /* start point bordered box */
  fltk::Group *scontour  = new fltk::Group(COLUMN1, 31, 240, 47, "Start point");
  scontour->box(fltk::BORDER_FRAME);
  scontour->align(fltk::ALIGN_INSIDE_LEFT);

  /* x1 */
  fltk::ValueInput *x1 = new fltk::ValueInput(COLUMN2, 34, 98, 20, "x: ");
  x1->range(-1000, 1000);
  x1->step(0.01);
  x1->linesize(0.01);
  inputindex[2] = boundary_subwindow->find(x1);

  /* y1 */
  fltk::ValueInput *y1 = new fltk::ValueInput(COLUMN2, 55, 98, 20, "y: ");
  y1->range(-1000, 1000);
  y1->step(0.01);
  y1->linesize(0.01);
  inputindex[3] = boundary_subwindow->find(y1);
  //  printf("%i", inputindex[2]);


  /* end point bordered box */
  fltk::Group *econtour  = new fltk::Group(COLUMN3, 31, 240, 47, "End point");
  econtour->box(fltk::BORDER_FRAME);
  econtour->align(fltk::ALIGN_INSIDE_LEFT);

  /* x2 */
  fltk::ValueInput *x2 = new fltk::ValueInput(COLUMN4, 34, 98, 20, "x: ");
  x2->range(-1000, 1000);
  x2->step(0.01);
  x2->linesize(0.01);
  //  x2->pagesize(1.0);
  inputindex[4] = boundary_subwindow->find(x2);
  //  printf("%i", inputindex[3]);

  /* y2 */
  fltk::ValueInput *y2 = new fltk::ValueInput(COLUMN4, 55, 98, 20, "y: ");
  y2->range(-1000, 1000);
  y2->step(0.01);
  y2->linesize(0.01);
  //  y2->pagesize(1.0);
  inputindex[5] = boundary_subwindow->find(y2);

  /* boundary */
  //  fltk::CheckButton *isboundary = new fltk::CheckButton(200, 70, 100, 20, "Boundary: ");
  //  isboundary->align(fltk::ALIGN_LEFT);
  //  inputindex[6] = boundary_subwindow->find(isboundary);
  /* Datagroup for [i|s|h]data */
  /*  fltk::Group *databox = new fltk::Group(10, 130, 580, 160, "");
  databox->box(fltk::BORDER_FRAME);
  */
  fltk::Button *new_dataline = new fltk::Button(10, 100, 580, 20, "Add data entry");
  new_dataline->callback((fltk::Callback *)new_dataline_cb);
  //  datagroup = new fltk::ScrollGroup(10, 140, 580, 160, "Insert [i|s|h]data here:");
  datagroup = new fltk::ScrollGroup(10, 130, 580, 160, "");
  datagroup->scrollbar.linesize(20);
  new_dataline_cb();
  new_dataline_cb();
  ((fltk::ValueInput *)(datagroup->child(1)))->set_value(0);
  ((fltk::ValueInput *)(datagroup->child(2)))->set_value(0.05);
  ((fltk::ValueInput *)(datagroup->child(4)))->set_value(1);
  ((fltk::ValueInput *)(datagroup->child(5)))->set_value(0.05);
  //  datagroup->box(fltk::BORDER_FRAME);
  //  datagroup->begin();
  //  fltk::Widget *box2 = new fltk::Widget(0, 0, 620, 160, "TEXT"); // this is to see the space avaiable
  //  fltk::Scroll *datascroll = new fltk::Scroll(600, 0, 20, 160, 0);
  //  datagroup->end();

  /* Buttons (Add, Cancel)*/
  fltk::Button *cancel = new fltk::Button(200, 305, 100, 25, "Cancel");
  cancel->image((fltk::Image *)cancel_icon);
  cancel->tooltip("Cancel");
  cancel->callback((fltk::Callback *)clear_cb);
  fltk::Button *add = new fltk::Button(320, 305, 100, 25, "&Add");
  add->image(LOAD_ICON_IMAGE("images/add.png"));
  cancel->tooltip("Add the primitive");
  add->callback((fltk::Callback *)add_line_cb);
  boundary_subwindow->end();
  boundary_subwindow->redraw();

  /* CHECKING AND FLUSHING */
#ifdef DEBUGGING
  printf("Selected line primitive\n");
  fflush(stdout);
#endif
}

void circle_cb(){
  /* POPULATE THE BOUNDARY_SUBWINDOW */
  clear_cb();
  circleitem->box(fltk::DOWN_BOX);
  boundary_subwindow->begin();
  /* spline */
  fltk::ValueInput *splinepoints = new fltk::ValueInput(COLUMN2, 10, 100, 20, "N. points for spline aprox.: ");
  splinepoints->range(0, 10000);
  splinepoints->linesize(1.0);
  splinepoints->step(1);
  splinepoints->set_value(100);
  inputindex[0] = boundary_subwindow->find(splinepoints);

  /* name */
  fltk::Input *primname = new fltk::Input(COLUMN4, 10, 100, 20, "Name (Optional): ");
  inputindex[1] = boundary_subwindow->find(primname);

  /* radius */
  fltk::ValueInput *radius = new fltk::ValueInput(COLUMN2, 31, 100, 20, "Radius: ");
  radius->range(-1000, 1000);
  radius->step(0.01);
  inputindex[2] = boundary_subwindow->find(radius);

  /* center point bordered box */
  fltk::Group *ccontour  = new fltk::Group(COLUMN1, 52, 240, 47, "Center");
  ccontour->box(fltk::BORDER_FRAME);
  ccontour->align(fltk::ALIGN_INSIDE_LEFT);

  /* center x */
  fltk::ValueInput *centerx = new fltk::ValueInput(COLUMN2, 55, 98, 20, "abscissa (x): ");
  centerx->range(-1000, 1000);
  centerx->step(0.01);
  centerx->linesize(0.01);
  inputindex[3] = boundary_subwindow->find(centerx);

  /* center y */
  fltk::ValueInput *centery = new fltk::ValueInput(COLUMN2, 76, 98, 20, "ordinata (y): ");
  centery->range(-1000, 1000);
  centery->step(0.01);
  centery->linesize(0.01);
  inputindex[4] = boundary_subwindow->find(centery);


  /* end point bordered box */
  fltk::Group *acontour  = new fltk::Group(COLUMN3, 52, 240, 47, "Angle (deg)");
  acontour->box(fltk::BORDER_FRAME);
  acontour->align(fltk::ALIGN_INSIDE_LEFT);

  /* start arc */
  fltk::ValueInput *sarc = new fltk::ValueInput(COLUMN4, 55, 98, 20, "start:");
  sarc->range(-360, 360);
  sarc->linesize(1.0);
  sarc->step(0.1);
  inputindex[5] = boundary_subwindow->find(sarc);

  /* end arc */
  fltk::ValueInput *earc = new fltk::ValueInput(COLUMN4, 76, 98, 20, "end:");
  earc->range(-360, 360);
  earc->linesize(1.0);
  earc->step(0.1);
  inputindex[6] = boundary_subwindow->find(earc);

  /* is boundary */
  //  fltk::CheckButton *isboundary = new fltk::CheckButton(200, 90, 100, 20, "Boundary: ");
  //  isboundary->align(fltk::ALIGN_LEFT);
  //  inputindex[7] = boundary_subwindow->find(isboundary);
  /* [i|s|h]data */
  fltk::Button *new_dataline = new fltk::Button(10, 120, 580, 20, "Add data entry");
  new_dataline->callback((fltk::Callback *)new_dataline_cb);
  datagroup = new fltk::ScrollGroup(10, 150, 580, 140, "");
  datagroup->scrollbar.linesize(20);
  new_dataline_cb();
  new_dataline_cb();
  ((fltk::ValueInput *)(datagroup->child(1)))->set_value(0);
  ((fltk::ValueInput *)(datagroup->child(2)))->set_value(0.05);
  ((fltk::ValueInput *)(datagroup->child(4)))->set_value(1);
  ((fltk::ValueInput *)(datagroup->child(5)))->set_value(0.05);
  /* buttons */
  fltk::Button *cancel = new fltk::Button(200, 305, 100, 25, "Cancel");
  cancel->image((fltk::Image *)cancel_icon);
  cancel->tooltip("Cancel");
  cancel->callback((fltk::Callback *)clear_cb);
  fltk::Button *add = new fltk::Button(320, 305, 100, 25, "Add");
  add->image(LOAD_ICON_IMAGE("images/add.png"));
  cancel->tooltip("Add the primitive");
  add->callback((fltk::Callback *)add_circle_cb);
  boundary_subwindow->end();
  boundary_subwindow->redraw();
  //  boundary_subwindow->redraw(0, 0, 620, 380);

  /* CHECKING AND FLUSHING */
#ifdef DEBUGGING
  printf("Selected circle primitive\n");
  fflush(stdout);
#endif
}

void ellipse_cb(){
  /* POPULATE THE BOUNDARY_SUBWINDOW */
  clear_cb();
  ellipseitem->box(fltk::DOWN_BOX);
  boundary_subwindow->begin();

  /* spline */
  fltk::ValueInput *splinepoints = new fltk::ValueInput(COLUMN2, 10, 100, 20, "N. points for spline aprox.: ");
  splinepoints->range(0, 10000);
  splinepoints->linesize(1.0);
  splinepoints->step(1);
  splinepoints->set_value(100);
  inputindex[0] = boundary_subwindow->find(splinepoints);

  /* name */
  fltk::Input *primname = new fltk::Input(COLUMN4, 10, 100, 20, "Name (Optional): ");
  inputindex[1] = boundary_subwindow->find(primname);

  /* Semixais Length */
  fltk::Group *scontour  = new fltk::Group(COLUMN1, 31, 240, 47, "Semiaxis length");
  scontour->box(fltk::BORDER_FRAME);
  scontour->align(fltk::ALIGN_INSIDE_LEFT);

  /* semix */
  fltk::ValueInput *semix = new fltk::ValueInput(COLUMN2, 34, 98, 20, "x:");
  semix->range(-1000, 1000);
  semix->step(0.01);
  inputindex[2] = boundary_subwindow->find(semix);

  /* semiy */
  fltk::ValueInput *semiy = new fltk::ValueInput(COLUMN2, 55, 98, 20, "y:");
  semiy->range(-1000, 1000);
  semiy->step(0.01);
  inputindex[3] = boundary_subwindow->find(semiy);

  /* center point bordered box */
  fltk::Group *ccontour  = new fltk::Group(COLUMN1, 80, 240, 47, "Center");
  ccontour->box(fltk::BORDER_FRAME);
  ccontour->align(fltk::ALIGN_INSIDE_LEFT);

  /* center x */
  fltk::ValueInput *centerx = new fltk::ValueInput(COLUMN2, 83, 98, 20, "abscissa (x): ");
  centerx->range(-1000, 1000);
  centerx->step(0.01);
  centerx->linesize(0.01);
  inputindex[4] = boundary_subwindow->find(centerx);

  /* center y */
  fltk::ValueInput *centery = new fltk::ValueInput(COLUMN2, 104, 98, 20, "ordinata (y): ");
  centery->range(-1000, 1000);
  centery->step(0.01);
  centery->linesize(0.01);
  inputindex[5] = boundary_subwindow->find(centery);


  /* Angle */
  fltk::Group *acontour  = new fltk::Group(COLUMN3, 31, 240, 47, "Angle (deg)");
  acontour->box(fltk::BORDER_FRAME);
  acontour->align(fltk::ALIGN_INSIDE_LEFT);

  /* start arc */
  fltk::ValueInput *sarc = new fltk::ValueInput(COLUMN4, 34, 98, 20, "start:");
  sarc->range(-360, 360);
  sarc->linesize(1.0);
  sarc->step(0.1);
  inputindex[6] = boundary_subwindow->find(sarc);

  /* end arc */
  fltk::ValueInput *earc = new fltk::ValueInput(COLUMN4, 55, 98, 20, "end:");
  earc->range(-360, 360);
  earc->linesize(1.0);
  earc->step(0.1);
  inputindex[7] = boundary_subwindow->find(earc);

  /* is boundary */
  //  fltk::CheckButton *isboundary = new fltk::CheckButton(200, 90, 100, 20, "Boundary: ");
  //  isboundary->align(fltk::ALIGN_LEFT);
  //  inputindex[8] = boundary_subwindow->find(isboundary);
  /* [i|s|h]data */
  fltk::Button *new_dataline = new fltk::Button(10, 140, 580, 20, "Add data entry");
  new_dataline->callback((fltk::Callback *)new_dataline_cb);
  datagroup = new fltk::ScrollGroup(10, 170, 580, 120, "");
  datagroup->scrollbar.linesize(20);
  new_dataline_cb();
  new_dataline_cb();
  ((fltk::ValueInput *)(datagroup->child(1)))->set_value(0);
  ((fltk::ValueInput *)(datagroup->child(2)))->set_value(0.05);
  ((fltk::ValueInput *)(datagroup->child(4)))->set_value(1);
  ((fltk::ValueInput *)(datagroup->child(5)))->set_value(0.05);
  /* buttons */
  fltk::Button *cancel = new fltk::Button(200, 305, 100, 25, "Cancel");
  cancel->image((fltk::Image *)cancel_icon);
  cancel->tooltip("Cancel");
  cancel->callback((fltk::Callback *)clear_cb);
  fltk::Button *add = new fltk::Button(320, 305, 100, 25, "Add");
  add->image(LOAD_ICON_IMAGE("images/add.png"));
  cancel->tooltip("Add the primitive");
  add->callback((fltk::Callback *)add_ellipse_cb);
  boundary_subwindow->end();
  boundary_subwindow->redraw();
  //  boundary_subwindow->redraw(0, 0, 620, 380);

  /* CHECKING AND FLUSHING */
#ifdef DEBUGGING
  printf("Selected ellipse primitive\n");
  fflush(stdout);
#endif
}

void data_cb(){
  /* POPULATE THE BOUNDARY_SUBWINDOW */
  clear_cb();
  dataitem->box(fltk::DOWN_BOX);
  boundary_subwindow->begin();

  /* rotation */
  fltk::ValueInput *rotation = new fltk::ValueInput(190, 10, 100, 20, "Rotation angle: ");
  rotation->range(-360, 360);
  rotation->linesize(1.0);
  rotation->step(1);
  rotation->set_value(0);
  inputindex[0] = boundary_subwindow->find(rotation);

  /* name */
  fltk::Input *primname = new fltk::Input(450, 10, 100, 20, "Name (Optional): ");
  inputindex[1] = boundary_subwindow->find(primname);


  /* start point bordered box */
  fltk::Group *scontour  = new fltk::Group(50, 31, 240, 47, "Displacement ");
  scontour->box(fltk::BORDER_FRAME);
  scontour->align(fltk::ALIGN_INSIDE_LEFT);

  /* trasx */
  fltk::ValueInput *trasx = new fltk::ValueInput(190, 34, 98, 20, "x: ");
  trasx->range(-1000, 1000);
  trasx->step(0.01);
  trasx->linesize(0.01);
  inputindex[2] = boundary_subwindow->find(trasx);

  /* trasy */
  fltk::ValueInput *trasy = new fltk::ValueInput(190, 55, 98, 20, "y: ");
  trasy->range(-1000, 1000);
  trasy->step(0.01);
  trasy->linesize(0.01);
  inputindex[3] = boundary_subwindow->find(trasy);

  /* data filename */
  fltk::Input *datafilename = new fltk::Input(COLUMN2, 80, 250, 20, "Filename :");
  inputindex[4] = boundary_subwindow->find(datafilename);

  /* Datagroup for [i|s|h]data */
  fltk::Button *new_dataline = new fltk::Button(10, 120, 580, 20, "Add data entry");
  new_dataline->callback((fltk::Callback *)new_dataline_cb);
  datagroup = new fltk::ScrollGroup(10, 150, 580, 140, "");
  datagroup->scrollbar.linesize(20);
  new_dataline_cb();
  new_dataline_cb();
  ((fltk::ValueInput *)(datagroup->child(1)))->set_value(0);
  ((fltk::ValueInput *)(datagroup->child(2)))->set_value(0.05);
  ((fltk::ValueInput *)(datagroup->child(4)))->set_value(1);
  ((fltk::ValueInput *)(datagroup->child(5)))->set_value(0.05);

  /* Buttons (Add, Cancel)*/
  fltk::Button *cancel = new fltk::Button(200, 305, 100, 25, "Cancel");
  cancel->image((fltk::Image *)cancel_icon);
  cancel->tooltip("Cancel");
  cancel->callback((fltk::Callback *)clear_cb);
  fltk::Button *add = new fltk::Button(320, 305, 100, 25, "Load");
  add->image(LOAD_ICON_IMAGE("images/add.png"));
  cancel->tooltip("Add the primitive");
  add->callback((fltk::Callback *)add_data_cb);
  boundary_subwindow->end();
  boundary_subwindow->redraw();

  /* CHECKING AND FLUSHING */
#ifdef DEBUGGING
  printf("Selected data\n");
  fflush(stdout);
#endif

}

void rectangle_cb(){
  /* POPULATE THE BOUNDARY_SUBWINDOW */
  clear_cb();
  rectangleitem->box(fltk::DOWN_BOX);
  boundary_subwindow->begin();

  /* spline*/
  fltk::ValueInput *splinepoints = new fltk::ValueInput(COLUMN2, 10, 100, 20, "N. points for spline aprox.: ");
  splinepoints->range(0, 10000);
  splinepoints->linesize(1.0);
  splinepoints->step(1);
  splinepoints->set_value(100);
  inputindex[0] = boundary_subwindow->find(splinepoints);

  /* name */
  fltk::Input *primname = new fltk::Input(COLUMN4, 10, 100, 20, "Name (Optional): ");
  inputindex[1] = boundary_subwindow->find(primname);

  /* start point bordered box */
  fltk::Group *scontour  = new fltk::Group(COLUMN1, 31, 240, 47, "Bottom left point");
  scontour->box(fltk::BORDER_FRAME);
  scontour->align(fltk::ALIGN_INSIDE_LEFT);

  /* x1 */
  fltk::ValueInput *x1 = new fltk::ValueInput(COLUMN2, 34, 98, 20, "x: ");
  x1->range(-1000, 1000);
  x1->step(0.01);
  x1->linesize(0.01);
  inputindex[2] = boundary_subwindow->find(x1);

  /* y1 */
  fltk::ValueInput *y1 = new fltk::ValueInput(COLUMN2, 55, 98, 20, "y: ");
  y1->range(-1000, 1000);
  y1->step(0.01);
  y1->linesize(0.01);
  inputindex[3] = boundary_subwindow->find(y1);


  /* end point bordered box */
  fltk::Group *econtour  = new fltk::Group(COLUMN3, 31, 240, 47, "Top right point");
  econtour->box(fltk::BORDER_FRAME);
  econtour->align(fltk::ALIGN_INSIDE_LEFT);

  /* x2 */
  fltk::ValueInput *x2 = new fltk::ValueInput(COLUMN4, 34, 98, 20, "x: ");
  x2->range(-1000, 1000);
  x2->step(0.01);
  x2->linesize(0.01);
  inputindex[4] = boundary_subwindow->find(x2);

  /* y2 */
  fltk::ValueInput *y2 = new fltk::ValueInput(COLUMN4, 55, 98, 20, "y: ");
  y2->range(-1000, 1000);
  y2->step(0.01);
  y2->linesize(0.01);
  inputindex[5] = boundary_subwindow->find(y2);

  /* Datagroup for [i|s|h]data */
  /*  fltk::Group *databox = new fltk::Group(10, 130, 580, 160, "");
  databox->box(fltk::BORDER_FRAME);
  */
  fltk::Button *new_dataline = new fltk::Button(10, 100, 580, 20, "Add data entry");
  new_dataline->callback((fltk::Callback *)new_dataline_cb);
  //  datagroup = new fltk::ScrollGroup(10, 140, 580, 160, "Insert [i|s|h]data here:");
  datagroup = new fltk::ScrollGroup(10, 130, 580, 160, "");
  datagroup->scrollbar.linesize(20);
  new_dataline_cb();
  new_dataline_cb();
  ((fltk::ValueInput *)(datagroup->child(1)))->set_value(0);
  ((fltk::ValueInput *)(datagroup->child(2)))->set_value(0.05);
  ((fltk::ValueInput *)(datagroup->child(4)))->set_value(1);
  ((fltk::ValueInput *)(datagroup->child(5)))->set_value(0.05);
  //  datagroup->box(fltk::BORDER_FRAME);
  //  datagroup->begin();
  //  fltk::Widget *box2 = new fltk::Widget(0, 0, 620, 160, "TEXT"); // this is to see the space avaiable
  //  fltk::Scroll *datascroll = new fltk::Scroll(600, 0, 20, 160, 0);
  //  datagroup->end();

  /* Buttons (Add, Cancel)*/
  fltk::Button *cancel = new fltk::Button(200, 305, 100, 25, "Cancel");
  cancel->image((fltk::Image *)cancel_icon);
  cancel->tooltip("Cancel");
  cancel->callback((fltk::Callback *)clear_cb);
  fltk::Button *add = new fltk::Button(320, 305, 100, 25, "&Add");
  add->image(LOAD_ICON_IMAGE("images/add.png"));
  cancel->tooltip("Add the primitive");
  add->callback((fltk::Callback *)add_rectangle_cb);
  boundary_subwindow->end();
  boundary_subwindow->redraw();

  /* CHECKING AND FLUSHING */
#ifdef DEBUGGING
  printf("Selected rectangle primitive\n");
  fflush(stdout);
#endif
}

void new_dataline_cb() {
  int currentpos;
  datagroup->begin();
  /* This is a hard hack since inputs (and any other thing) got created 
   * from actual origin and not from widget (datagroup in this case)
   * absolute position, anyway the first time the input got creted with
   * the bar fully scored an unexpected 10 pixel offset is generated.
   * This can be an initialization bug in fltk code.
   * I figured out the problem and resoleved it, it is a bad, but not
   * buggy, behavor or fltk. It just wipesout white space if not filled
   * by some widget (window or anything) */
  /*
  if(datagroup->yposition() < 0) {
    currentpos = ydataposition;
  }
  else {
    currentpos = ydataposition - datagroup->yposition();
  }
  */
  datagroup->yposition() < 0 ?\
    currentpos = ydataposition :\
    currentpos = ydataposition - datagroup->yposition();
  //  fprintf(stdout, "yposition(): %i\n", datagroup->yposition());
  //  fprintf(stdout, "currentpos: %i\n", currentpos);
  //  fflush(stdout);
  //  ydataposition = ydataposition - datagroup->yposition();
  fltk::ValueInput *idata = new fltk::ValueInput(50, currentpos, 100, 20, "idata: ");
  idata->range(0, 3);
  idata->linesize(1.0);
  idata->set_value(1);
  fltk::ValueInput *sdata = new fltk::ValueInput(250, currentpos, 100, 20, "sdata: ");
  sdata->range(0, 1);
  sdata->linesize(0.01);
  sdata->set_value(1);
  fltk::ValueInput *hdata = new fltk::ValueInput(450, currentpos, 100, 20, "hdata: ");
  hdata->range(0, 1);
  hdata->linesize(0.01);
  hdata->set_value(0.05);
  datagroup->end();
  datagroup->redraw();
  //  datagroup->redraw(0, currentpos, 620, 20);  // maybe this is not anymore necessary, but should be correct.
  ydataposition = ydataposition + 20;
}

void clear_cb(){
  /* reset the scroll position */
  ydataposition = 0;
  /* clean the index for primitive values */
  for(int i = 0; i < 10; i++) inputindex[i] = 0;
  /* reset button state */
  lineitem->box(fltk::NO_BOX);
  circleitem->box(fltk::NO_BOX);
  ellipseitem->box(fltk::NO_BOX);
  dataitem->box(fltk::NO_BOX);
  rectangleitem->box(fltk::NO_BOX);
  /* clear the window */
  boundary_subwindow->clear();
  boundary_subwindow->redraw();
  //  boundary_subwindow->redraw(0, 0, 620, 330);
}

void add_line_cb() {
  int ndata;
  int validdata = 0;
  Line *prim = (Line *)(u2d_core->add_prim(PR_LINE));
  //  printf("boundary_subwindow ha %i figli\n", boundary_subwindow->children());
  //  strncpy(prim->primname, "ciao", 31);
  strncpy(prim->primname, ((fltk::Input *)boundary_subwindow->child(inputindex[1]))->value(), 32);
  /* Changed to use the new function
  prim->spline = (int)((fltk::ValueInput *)boundary_subwindow->child(inputindex[0]))->value();
  prim->x1 = ((fltk::ValueInput *)boundary_subwindow->child(inputindex[2]))->value();
  prim->y1 = ((fltk::ValueInput *)boundary_subwindow->child(inputindex[3]))->value();
  prim->x2 = ((fltk::ValueInput *)boundary_subwindow->child(inputindex[4]))->value();
  prim->y2 = ((fltk::ValueInput *)boundary_subwindow->child(inputindex[5]))->value();
  */
  prim->set_coord((int)((fltk::ValueInput *)boundary_subwindow->child(inputindex[0]))->value(),\
                  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[2]))->value(),\
		  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[3]))->value(),\
		  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[4]))->value(),\
		  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[5]))->value());
  printf("Adding line primitive\n");
  printf("Spline: %i x1: %f y1: %f x2: %f y2: %f\n", prim->spline, prim->x1, prim->y1, prim->x2, prim->y2);
  fflush(stdout);
  //  prim->isboundary = ((fltk::CheckButton *)boundary_subwindow->child(inputindex[6]))->value();
  //  printf("Boundary : %s\n",  (prim->isboundary) ? "true" : "false");
  ndata = datagroup->children();
  prim->allocatedata(ndata/3);
  /* this could be done nicer with k%3 as second argument and some
   * other arragnements, but it's ok for now
   */
  for(int k = 0; k < ndata; k = k + 3) {
    if(((fltk::ValueInput *)datagroup->child(k))->value() > 0) {
#ifdef DEBUGGING
      fprintf(stdout, "k/3: %i child n.: %i valid data (x): %i\n", k/3, k, validdata);
      fflush(stdout);
#endif
      prim->setdata(validdata, 0, ((fltk::ValueInput *)datagroup->child(k))->value());
      prim->setdata(validdata, 1, ((fltk::ValueInput *)datagroup->child(k + 1))->value());
      prim->setdata(validdata, 2, ((fltk::ValueInput *)datagroup->child(k + 2))->value());
      validdata++;
      prim->curdata++;
    }
  }
  add_generic(prim);
}

void add_circle_cb() {
  int ndata;
  int validdata = 0;
  Circle *prim = (Circle *)(u2d_core->add_prim(PR_CIRCLE));
#ifdef DEBUGGING
  printf("boundary_subwindow ha %i figli\n", boundary_subwindow->children());
#endif
  strncpy(prim->primname, ((fltk::Input *)boundary_subwindow->child(inputindex[1]))->value(), 32);
  /* Changed to use the new function
  prim->spline = (int)((fltk::ValueInput *)boundary_subwindow->child(inputindex[0]))->value();
  prim->radius = ((fltk::ValueInput *)boundary_subwindow->child(inputindex[2]))->value();
  prim->centerx = ((fltk::ValueInput *)boundary_subwindow->child(inputindex[3]))->value();
  prim->centery = ((fltk::ValueInput *)boundary_subwindow->child(inputindex[4]))->value();
  prim->startarc = ((fltk::ValueInput *)boundary_subwindow->child(inputindex[5]))->value();
  prim->endarc = ((fltk::ValueInput *)boundary_subwindow->child(inputindex[6]))->value();
  */
  prim->set_coord((int)((fltk::ValueInput *)boundary_subwindow->child(inputindex[0]))->value(),/* spline */ \
    	  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[2]))->value(), /* radius */ \
    	  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[3]))->value(), /* centerx */ \
    	  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[4]))->value(), /* centery */ \
    	  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[5]))->value(), /* startarc */ \
    	  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[6]))->value()); /* endarc */
  printf("Adding circle primitive\n");
  printf("Spline: %i radius: %f centerx: %f centery: %f startarc: %f endarc: %f\n", prim->spline, prim->radius, prim->centerx, prim->centery, prim->startarc, prim->endarc);
  fflush(stdout);
  //  prim->isboundary = ((fltk::CheckButton *)boundary_subwindow->child(inputindex[7]))->value();
  ndata = datagroup->children();
  prim->allocatedata(ndata/3);
  /* This for cycle could be done nicer with k%3 as second argument and some
   * other arragnements, but it's ok for now
   */
  for(int k = 0; k < ndata; k = k + 3) {
    if(((fltk::ValueInput *)datagroup->child(k))->value() > 0) {
#ifdef DEBUGGING
      fprintf(stdout, "k/3: %i child n.: %i valid data (x): %i\n", k/3, k, validdata);
      fflush(stdout);
#endif
      prim->setdata(validdata, 0, ((fltk::ValueInput *)datagroup->child(k))->value());
      prim->setdata(validdata, 1, ((fltk::ValueInput *)datagroup->child(k + 1))->value());
      prim->setdata(validdata, 2, ((fltk::ValueInput *)datagroup->child(k + 2))->value());
      validdata++;
      prim->curdata++;
    }
  }
  add_generic(prim);
}

void add_ellipse_cb() {
  int ndata;
  int validdata = 0;
  Ellipse *prim = (Ellipse *)(u2d_core->add_prim(PR_ELLIPSE));
#ifdef DEBUGGING
  printf("boundary_subwindow ha %i figli\n", boundary_subwindow->children());
#endif
  strncpy(prim->primname, ((fltk::Input *)boundary_subwindow->child(inputindex[1]))->value(), 32);
  /* Changed to use the new function
  prim->spline = (int)((fltk::ValueInput *)boundary_subwindow->child(inputindex[0]))->value();
  prim->semix = ((fltk::ValueInput *)boundary_subwindow->child(inputindex[2]))->value();
  prim->semiy = ((fltk::ValueInput *)boundary_subwindow->child(inputindex[3]))->value();
  prim->centerx = ((fltk::ValueInput *)boundary_subwindow->child(inputindex[4]))->value();
  prim->centery = ((fltk::ValueInput *)boundary_subwindow->child(inputindex[5]))->value();
  prim->startarc = ((fltk::ValueInput *)boundary_subwindow->child(inputindex[6]))->value();
  prim->endarc = ((fltk::ValueInput *)boundary_subwindow->child(inputindex[7]))->value();
  */
  prim->set_coord((int)((fltk::ValueInput *)boundary_subwindow->child(inputindex[0]))->value(),  /* spline */ \
  	  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[2]))->value(), /* semix */ \
    	  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[3]))->value(), /* semiy */ \
    	  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[4]))->value(), /* centerx */ \
    	  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[5]))->value(), /* centery */ \
    	  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[6]))->value(), /* startarc */ \
    	  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[7]))->value()); /* endarc */
  printf("Adding ellipse primitive\n");
  printf("Spline: %i semix: %f semiy: %f centerx: %f centery: %f startarc: %f endarc: %f\n", prim->spline, prim->semix, prim->semiy, prim->centerx, prim->centery, prim->startarc, prim->endarc);
  fflush(stdout);
  //  prim->isboundary = ((fltk::CheckButton *)boundary_subwindow->child(inputindex[8]))->value();
  ndata = datagroup->children();
  prim->allocatedata(ndata/3);
  /* This could be done nicer with k%3 as second argument and some
   * other arragnements, but it's ok for now
   */
  for(int k = 0; k < ndata; k = k + 3) {
    if(((fltk::ValueInput *)datagroup->child(k))->value() > 0) {
#ifdef DEBUGGING
      fprintf(stdout, "k/3: %i child n.: %i valid data (x): %i\n", k/3, k, validdata);
      fflush(stdout);
#endif
      prim->setdata(validdata, 0, ((fltk::ValueInput *)datagroup->child(k))->value());
      prim->setdata(validdata, 1, ((fltk::ValueInput *)datagroup->child(k + 1))->value());
      prim->setdata(validdata, 2, ((fltk::ValueInput *)datagroup->child(k + 2))->value());
      validdata++;
      prim->curdata++;
    }
  }
  add_generic(prim);
}

void add_data_cb(){
  int ndata;
  int validdata = 0;
  int retval;
  int path_max;

  printf("Adding data primitive\n");

  Data *prim = (Data *)(u2d_core->add_prim(PR_DATA));
#ifdef DEBUGGING
  printf("boundary_subwindow ha %i figli\n", boundary_subwindow->children());
#endif
  strncpy(prim->primname, ((fltk::Input *)boundary_subwindow->child(inputindex[1]))->value(), 32);

  /* expand filename */
#ifdef PATH_MAX
  path_max = PATH_MAX;
#else
  path_max = pathconf (((fltk::Input *)boundary_subwindow->child(inputindex[4]))->value(), _PC_PATH_MAX);
  if (path_max <= 0)
    path_max = 4096;
#endif
  char *filename_expanded = NULL;
  filename_expanded = (char *)malloc(path_max * sizeof(char));
  if(filename_expanded == NULL){
    fprintf(stderr, "Cannot allocate memory for pathname expansion\n");
    msgout("ERROR: Cannot allocate memory for pathname expansion");
    u2d_core->del_prim(prim);
    return;
  }
  memset(filename_expanded, 0, path_max);
  realpath(((fltk::Input *)boundary_subwindow->child(inputindex[4]))->value(),\
	   filename_expanded);
  /*
    This was and is bad code.. removed now, here for pasting propouses
  if(strlen(filename_expanded) >= length) {
    fprintf(stderr, "add_data_cb: Filename is too long, abroting.");
    msgout("ERROR: Filename and path too long");
    fltk::alert("Filename and path too long, cannot parse data file\n Please move it in a directory near to root\n");
    u2d_core->del_prim(prim);
    return;
  }
  */
#ifdef DEBUGGING
  printf("add_data_cb: Filename is: %s\n", filename_expanded);
#endif
  retval = prim->set_coord(filename_expanded,  /* filename */ \
			   ((fltk::ValueInput *)boundary_subwindow->child(inputindex[2]))->value(), /* trasx */ \
			   ((fltk::ValueInput *)boundary_subwindow->child(inputindex[3]))->value(), /* trasy */ \
			   ((fltk::ValueInput *)boundary_subwindow->child(inputindex[0]))->value()); /* rotation */
  free(filename_expanded);
  if(retval < 0) {
    msgout("ERROR loading data file");
    if(retval == NO_FILE) fltk::alert("Error opening file: Invalid data filename\n");
    else if(retval == INVALID_FILE) fltk::alert("Error reading file: check file syntax\n");
    else if(retval == NOT_A_FILE) fltk::alert("Error with file: not a regular file\n");
    else if(retval == NO_EXIST) fltk::alert("Error with filename: check filename syntax\n");
    else fltk::alert("Error opening file: Unkown reason\n");
    u2d_core->del_prim(prim);
    return;
  }
  msgout("Data file loaded, acquired %i points", prim->npoints);
  printf("Filename: %s trasx: %f trasy: %f rotation: %f\n",\
	 prim->filename, prim->trasx, prim->trasy,\
	 prim->rotation);
  fflush(stdout);
  ndata = datagroup->children();
  prim->allocatedata(ndata/3);
  /* This could be done nicer with k%3 as second argument and some
   * other arragnements, but it's ok for now
   */
  for(int k = 0; k < ndata; k = k + 3) {
    if(((fltk::ValueInput *)datagroup->child(k))->value() > 0) {
#ifdef DEBUGGING
      fprintf(stdout, "k/3: %i child n.: %i valid data (x): %i\n", k/3, k, validdata);
      fflush(stdout);
#endif
      prim->setdata(validdata, 0, ((fltk::ValueInput *)datagroup->child(k))->value());
      prim->setdata(validdata, 1, ((fltk::ValueInput *)datagroup->child(k + 1))->value());
      prim->setdata(validdata, 2, ((fltk::ValueInput *)datagroup->child(k + 2))->value());
      validdata++;
      prim->curdata++;
    }
  }
  add_generic(prim);

}

void add_rectangle_cb() {


/* PLACEHOLDER TO CODE */

/* Bottom left -> Top left */
  int ndata;
  int validdata = 0;
  Line *prim = (Line *)(u2d_core->add_prim(PR_LINE));
  strncpy(prim->primname, ((fltk::Input *)boundary_subwindow->child(inputindex[1]))->value(), 32);
  prim->set_coord((int)((fltk::ValueInput *)boundary_subwindow->child(inputindex[0]))->value(),\
                  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[2]))->value(),\
		  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[3]))->value(),\
		  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[2]))->value(),\
		  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[5]))->value());
  printf("Adding rectangle primitive\n");
  printf("Spline: %i x1: %f y1: %f x2: %f y2: %f\n", prim->spline, prim->x1, prim->y1, prim->x2, prim->y2);
  fflush(stdout);

  ndata = datagroup->children();
  prim->allocatedata(ndata/3);
  /* this could be done nicer with k%3 as second argument and some
   * other arragnements, but it's ok for now
   */
  for(int k = 0; k < ndata; k = k + 3) {
    if(((fltk::ValueInput *)datagroup->child(k))->value() > 0) {
#ifdef DEBUGGING
      fprintf(stdout, "k/3: %i child n.: %i valid data (x): %i\n", k/3, k, validdata);
      fflush(stdout);
#endif
      prim->setdata(validdata, 0, ((fltk::ValueInput *)datagroup->child(k))->value());
      prim->setdata(validdata, 1, ((fltk::ValueInput *)datagroup->child(k + 1))->value());
      prim->setdata(validdata, 2, ((fltk::ValueInput *)datagroup->child(k + 2))->value());
      validdata++;
      prim->curdata++;
    }
  }
  add_generic(prim);

/* Top left -> Top right */
  validdata = 0;
  prim = (Line *)(u2d_core->add_prim(PR_LINE));
  strncpy(prim->primname, ((fltk::Input *)boundary_subwindow->child(inputindex[1]))->value(), 32);
  prim->set_coord((int)((fltk::ValueInput *)boundary_subwindow->child(inputindex[0]))->value(),\
                  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[2]))->value(),\
		  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[5]))->value(),\
		  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[4]))->value(),\
		  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[5]))->value());
  printf("Spline: %i x1: %f y1: %f x2: %f y2: %f\n", prim->spline, prim->x1, prim->y1, prim->x2, prim->y2);
  fflush(stdout);

  prim->allocatedata(ndata/3);
  /* this could be done nicer with k%3 as second argument and some
   * other arragnements, but it's ok for now
   */
  for(int k = 0; k < ndata; k = k + 3) {
    if(((fltk::ValueInput *)datagroup->child(k))->value() > 0) {
#ifdef DEBUGGING
      fprintf(stdout, "k/3: %i child n.: %i valid data (x): %i\n", k/3, k, validdata);
      fflush(stdout);
#endif
      prim->setdata(validdata, 0, ((fltk::ValueInput *)datagroup->child(k))->value());
      prim->setdata(validdata, 1, ((fltk::ValueInput *)datagroup->child(k + 1))->value());
      prim->setdata(validdata, 2, ((fltk::ValueInput *)datagroup->child(k + 2))->value());
      validdata++;
      prim->curdata++;
    }
  }
  add_generic(prim);

/* Top right -> Bottom right */
  validdata = 0;
  prim = (Line *)(u2d_core->add_prim(PR_LINE));
  strncpy(prim->primname, ((fltk::Input *)boundary_subwindow->child(inputindex[1]))->value(), 32);
  prim->set_coord((int)((fltk::ValueInput *)boundary_subwindow->child(inputindex[0]))->value(),\
                  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[4]))->value(),\
		  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[5]))->value(),\
		  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[4]))->value(),\
		  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[3]))->value());
  printf("Spline: %i x1: %f y1: %f x2: %f y2: %f\n", prim->spline, prim->x1, prim->y1, prim->x2, prim->y2);
  fflush(stdout);

  prim->allocatedata(ndata/3);
  /* this could be done nicer with k%3 as second argument and some
   * other arragnements, but it's ok for now
   */
  for(int k = 0; k < ndata; k = k + 3) {
    if(((fltk::ValueInput *)datagroup->child(k))->value() > 0) {
#ifdef DEBUGGING
      fprintf(stdout, "k/3: %i child n.: %i valid data (x): %i\n", k/3, k, validdata);
      fflush(stdout);
#endif
      prim->setdata(validdata, 0, ((fltk::ValueInput *)datagroup->child(k))->value());
      prim->setdata(validdata, 1, ((fltk::ValueInput *)datagroup->child(k + 1))->value());
      prim->setdata(validdata, 2, ((fltk::ValueInput *)datagroup->child(k + 2))->value());
      validdata++;
      prim->curdata++;
    }
  }
  add_generic(prim);

/* Bottom right -> Bottom left */
  validdata = 0;
  prim = (Line *)(u2d_core->add_prim(PR_LINE));
  strncpy(prim->primname, ((fltk::Input *)boundary_subwindow->child(inputindex[1]))->value(), 32);
  prim->set_coord((int)((fltk::ValueInput *)boundary_subwindow->child(inputindex[0]))->value(),\
                  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[4]))->value(),\
		  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[3]))->value(),\
		  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[2]))->value(),\
		  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[3]))->value());
  printf("Spline: %i x1: %f y1: %f x2: %f y2: %f\n", prim->spline, prim->x1, prim->y1, prim->x2, prim->y2);
  fflush(stdout);

  prim->allocatedata(ndata/3);
  /* this could be done nicer with k%3 as second argument and some
   * other arragnements, but it's ok for now
   */
  for(int k = 0; k < ndata; k = k + 3) {
    if(((fltk::ValueInput *)datagroup->child(k))->value() > 0) {
#ifdef DEBUGGING
      fprintf(stdout, "k/3: %i child n.: %i valid data (x): %i\n", k/3, k, validdata);
      fflush(stdout);
#endif
      prim->setdata(validdata, 0, ((fltk::ValueInput *)datagroup->child(k))->value());
      prim->setdata(validdata, 1, ((fltk::ValueInput *)datagroup->child(k + 1))->value());
      prim->setdata(validdata, 2, ((fltk::ValueInput *)datagroup->child(k + 2))->value());
      validdata++;
      prim->curdata++;
    }
  }
  add_generic(prim);

}

void add_generic(Primitive *prim){
  u2d_gl->set_prim_list_start(u2d_core->get_prim_list_start(), \
		              u2d_core->get_prim_number());
  u2d_gl->guess_ref_size(prim);
  u2d_gl->redraw();
  prim_browser->add(u2d_core->get_prim_string(prim), 0, NULL, 0, 0);
}

void openitem_cb(){
 fltk::message("Open Feature not avaialbe yet\n");
}

void saveasitem_cb(){
  openfilebrowser();
  creation_nextbutton_cb();
}

void saveitem_cb(){
  if(strlen(((fltk::Input *)creation_subwindow->child(filenameinput_i))->value()) < 1) saveasitem_cb();
  else creation_nextbutton_cb();
}

void quititem_cb(){
  window->hide();
  opengl_window->hide();
  prim_browser_window->hide();
  /* calling the upper 2 or this one is exacly the same, no idea what is better */
  /* exit(0); */
}

void about_cb(){
  fltk::message("%s by Gabriele Dini Ciacci - version %s\n", PACKAGE, VERSION);
}

void boundary_nextbutton_cb(){
  boundarygroup->hide();
  domaingroup->show();
  msgout("Define domain constraints");
}

void domain_nextbutton_cb(){
  domaingroup->hide();
  creationgroup->show();
  msgout("Define output options");
}

void domain_prevbutton_cb(){
  domaingroup->hide();
  boundarygroup->show();
  msgout("Back to boundary definition");
}

void creation_prevbutton_cb(){
  creationgroup->hide();
  domaingroup->show();
  msgout("Back to domain definition");
}

void umesh2d_format_cb(){
  fileformat = FORMAT_UMESH2D;
}

void umesh2d_fe_format_cb(){
  fileformat = FORMAT_UMESH2D_FE;
}

void umesh2d_fv_format_cb(){
  fileformat = FORMAT_UMESH2D_FV;
}


void glview_cb(){
  if(opengl_window->visible()){
    opengl_window->hide();
  }
  else {
    opengl_window->show();
    opengl_window->position(window->x() + window->w() + MANAGER_BORDER, window->y());
  }
  /* Removed becausenow we hide the main detached window
   * and not anymore the gl window itself */
  /*
  if(u2d_gl->visible()){
    u2d_gl->hide();
  }
  else {
    u2d_gl->show();
  }
  */
}

void browserview_cb(){
  if(prim_browser_window->visible()){
    prim_browser_window->hide();
  }
  else {
    prim_browser_window->show();
    prim_browser_window->position(window->x(), window->y() - BROWSER_H - TITLEBAR_OFFSET);
  }
}

void browser_del_prim_cb() {
  fltk::Widget* current = prim_browser->goto_top();
  while (current) {
    if (current->selected()) {
      /* This work in case of hierarical menus too.
       * In my case the current->parent() should allways be
       * prim_browser */
      fltk::Group* parent = current->parent();
      parent->remove(current);
      /* Now we  delete the primitive from the list like:
       * id = current->value() to get the removed primitive id
       * del_prim(get_prim_by_id(id)); */
      u2d_core->del_prim(u2d_core->get_prim_by_label(current->label()));
      delete current;
      parent->relayout();
      /* We start again from top to see if there are
       * still other primitives selected */
      current = prim_browser->goto_top();
    }
    else {
      current = prim_browser->next();
    }
  }

  /* We have to update the openGL window too */
  u2d_gl->set_prim_list_start(u2d_core->get_prim_list_start(), \
			      u2d_core->get_prim_number());
  /* We have to recalculate the reference sizes from
   * scratch... VERY costly */
  u2d_gl->guess_ref_size();
  /* We can finally redaw */
  u2d_gl->redraw();
}

void main_core_cb(){
  /* Calculate closed objects */
  fprintf(stdout, "Calculating closed objects and vertex enumeration...\n");
  u2d_core->find_objects();
  fprintf(stdout, "Done!\n");
}
 

void msgout(char *format, ...) {
  va_list ap;
  va_start(ap, format);
  int length = vsnprintf(msgbuf, MSGBUF_LENGTH, format, ap);
  memset(msgbuf + sizeof(char)*length + 1, 0, MSGBUF_LENGTH - sizeof(char)*length - 1);
  msgbar->value(msgbuf);
  va_end(ap);
}


void new_backgrid_cb() {
  int currentpos;
  backgridgroup->begin();
  /* This is a hard hack since inputs (and any other thing) got created 
   * from actual origin and not from widget (backgrid in this case)
   * absolute position, anyway the first time the input got creted with
   * the bar fully scored and unexpected 10 pixel offset is generated.
   * I suspect this is an initialization bug in fltk code.
   * I figured out the problem and resoleved it, it is a bad, but not
   * buggy, behavor or fltk. It just wipesout white space if not filled
   * by some widget (window or anything) */

  backgridgroup->yposition() < 0 ?\
    currentpos = ybackgridposition :\
    currentpos = ybackgridposition - backgridgroup->yposition();

  fltk::ValueInput *x = new fltk::ValueInput(40, currentpos, 100, 20, "x: ");
  x->linesize(0.01);
  fltk::ValueInput *y = new fltk::ValueInput(180, currentpos, 100, 20, "y: ");
  y->linesize(0.01);
  fltk::ValueInput *hdata = new fltk::ValueInput(330, currentpos, 100, 20, "hdata: ");
  hdata->linesize(0.01);
  fltk::Button *backgrid_del = new fltk::Button (440, currentpos, 80, 20, "delete");
  backgrid_del->callback((fltk::Callback *)backgrid_del_cb);
  backgridgroup->end();
  backgridgroup->redraw();
  ybackgridposition = ybackgridposition + 20;

}

void new_steiner_cb() {
  int currentpos;
  steinergroup->begin();
  /* This is a hard hack since inputs (and any other thing) got created 
   * from actual origin and not from widget (backgrid in this case)
   * absolute position, anyway the first time the input got creted with
   * the bar fully scored and unexpected 10 pixel offset is generated.
   * I suspect this is an initialization bug in fltk code.
   * I figured out the problem and resoleved it, it is a bad, but not
   * buggy, behavor or fltk. It just wipesout white space if not filled
   * by some widget (window or anything) */

  steinergroup->yposition() < 0 ?\
    currentpos = ysteinerposition :\
    currentpos = ysteinerposition - steinergroup->yposition();

  fltk::ValueInput *x = new fltk::ValueInput(40, currentpos, 100, 20, "x: ");
  x->linesize(0.01);
  fltk::ValueInput *y = new fltk::ValueInput(240, currentpos, 100, 20, "y: ");
  y->linesize(0.01);
  fltk::Button *steiner_del = new fltk::Button (440, currentpos, 80, 20, "delete");
  steiner_del->callback((fltk::Callback *)steiner_del_cb);
  steinergroup->end();
  steinergroup->redraw();
  ysteinerposition = ysteinerposition + 20;
}


void backgrid_del_cb(fltk::Widget *button, void*) {
  /* an integer that will contain the child id of the button */
  int button_line = 0;

  /* we get the parent, shoudlbe teh ScrollGroup */
  fltk::Group *parent = ((fltk::Button *)button)->parent();

  /* we find the button child index */
  button_line = parent->find(button);

  /* We delete the lines, notice thay we delete for 3 times the same index.
   * This is because the childs get renumberded after we delete one
   * so if we delete child n the n+1 child becomes himself the n child.
   */
  delete parent->child(button_line - 3);
  delete parent->child(button_line - 3);
  delete parent->child(button_line - 3);
  delete parent->child(button_line - 3);
}

void steiner_del_cb(fltk::Widget *button, void*) {
  /* an integer that will contain the child id of the button */
  int button_line = 0;

  /* we get the parent, shoudlbe teh ScrollGroup */
  fltk::Group *parent = ((fltk::Button *)button)->parent();

  /* we find the button child index */
  button_line = parent->find(button);

  /* We delete the lines, notice thay we delete for 3 times the same index.
   * This is because the childs get renumberded after we delete one
   * so if we delete child n the n+1 child becomes himself the n child.
   */
  delete parent->child(button_line - 2);
  delete parent->child(button_line - 2);
  delete parent->child(button_line - 2);
}


/* BUG we have  no way to know if we are in the correct dir if
 * the user name the last dir the same project dir like:
 * /home/name/mesh/mesh
 * This will cause confusion as if we are in /home/name/mesh
 * the program will not cd inside mesh.
 */
void create_dir() {
    char *buf = NULL;
    int path_size = 512;
    buf = (char *)malloc(path_size);
    char *current_path = NULL;
    bool unrec_error = false;

    while(current_path == NULL || unrec_error) {
        current_path = getcwd(buf, path_size);
        if(current_path == NULL && errno == ERANGE) {
            free(buf);
            path_size = path_size + 512;
            buf = (char *)malloc(path_size);
        }
        else if(current_path == NULL) {
            free(buf);
            unrec_error = true;
            printf("Cannot get path!\n");
            exit(-1);
        }
    }
    char *current_dir = rindex(current_path, '/') + 1;
    printf("%s\n", current_path);
    printf("%s\n", current_dir);

    int retval = 0;
    struct stat current_stat;
//    const char *project_dir = string(bdfile->get_project_name()).c_str();
    const char *project_dir = "test4";
    /* If we are not in the project dir enter it*/
    printf("project dir: %s\n", project_dir);
    if(current_dir != project_dir) {
        /* Check if the dir exist */
        retval = stat(project_dir, &current_stat);
        if(retval == -1 && errno == ENOENT) {
            printf("Directory does not exist, trying to create\n");
            /* Create the dir */
            retval = mkdir("test4", 0777);
            if(retval == -1) {
                /* unrecoverable error DO SOMETHING */
                /* Cannot create the dir */
                printf("Cannot create directory\n");
                return;
            }
            printf("Directory successfuly created\n");
            /* Now stat the created dir */
            retval = stat(project_dir, &current_stat);
        }
        else if(retval == -1) {
            /* unrecoverable error DO SOMETHING */
            /* Cannot stat the dir, but not because it's not exist */
            return;
        }

        if(S_ISDIR(current_stat.st_mode)) {
            /* directory already exists CD it */
            retval = chdir("test4");
            if(retval == -1) {
                printf("Cannot change to project directory\n");
            }
        }
        else {
            /* unrecoverable error DO SOMETHING */
            /* there is a file maybe with the same name as the dir */
            printf("Not a directory, cannot cd\n");
        }
    }

    free(buf);
    buf = NULL;
    path_size = 512;
    buf = (char *)malloc(path_size);
    current_path = NULL;
    unrec_error = false;
    while(current_path == NULL || unrec_error) {
        current_path = getcwd(buf, path_size);
        if(current_path == NULL && errno == ERANGE) {
            free(buf);
            path_size = path_size + 512;
            buf = (char *)malloc(path_size);
        }
        else if(current_path == NULL) {
            free(buf);
            unrec_error = true;
            printf("Cannot get path!\n");
            exit(-1);
        }
    }
    current_dir = rindex(current_path, '/') + 1;
    printf("current_path: %s\n", current_path);
    printf("current_dir: %s\n", current_dir);

}

void debug_cb(){
/*
    FILE *file = NULL;
    unsigned int npoints = 0;
    int val = 0;
    float *points;

    file = fopen("2dmesh", "r");
    if(file == NULL) {
        fprintf(stderr, "Cannot open file\n");
        return;
    }
    val = fscanf(file, "%d\n", &npoints);
    if(val != 1) {
        fprintf(stderr, "Cannot read the number of points in file\n");
        return;
    }
    points = (float *)calloc(npoints, 2*sizeof(float)); 
    for(unsigned int i = 1; i < npoints*2; i = i + 2) {
        val = fscanf(file, "  %f %f       0\n", (points + i - 1), (points + i));
        if(val != 2) {
            fprintf(stderr, "debug_cb: Cannot acquire correctly from"\
                    "file 2dmesh. Check line %i "\
                    "is in the correct format i.e. 0.034 0.23\n",\
                    i);
            fclose(file);
            return;
        }
        printf("Acquired point: (%f, %f)\n", *(points + i - 1), *(points + i));
    }
    fclose(file);

    Display_Set display_set;
    display_set.vertex_array = points;
    display_set.vertex_array_n = npoints;
    display_set.type = GL_LINES;
*/
//2>&1

    create_dir();
}

void execute_mesher() {
    string cmd_line = "echo " + string(bdfile->get_project_name()) + " | ./umesh2d";

    FILE *fp = popen(cmd_line.c_str(), "r");
    setvbuf(fp, NULL, _IOLBF, 256);

    char buffer[256];
    memset(buffer, 0, 256);
    while(fgets(buffer, 256, fp)) {
        printf("%s", buffer);
        shell_output->append(buffer);
    }
    pclose(fp);

}

void draw_mesh() {

    u2d_Load_PlotMtv dataloader(("gridplot." + string(bdfile->get_project_name())).c_str());
    u2d_glmesh->display_set.clear();
    u2d_glmesh->display_set.insert(u2d_glmesh->display_set.end(),\
                                   dataloader.display_set.begin(),\
                                   dataloader.display_set.end());
    u2d_glmesh->load_all_display_set();

}


/***************** TO CHECK AND REMOVE *******************/
void flush_output(FILE *fp, fltk::TextDisplay *shell_output) {
    char buffer[256];
    memset(buffer, 0, 256);
    while(fgets(buffer, 256, fp)) {
        printf("%s", buffer);
        shell_output->append(buffer);
    }
    pclose(fp);
    exit(0);
}
/********************************************************/
